import dataclasses
import typing
import weave
from .. import panel
from . import table_state

from .. import graph
from .. import weave_internal


@weave.type()
class FacetDimsConfig:
    x: str
    y: str
    select: str
    detail: str


@weave.type()
class FacetCellSize:
    w: int
    h: int


class FacetCell(typing.TypedDict):
    x: str
    y: str


@weave.type()
class FacetConfig:
    table: table_state.TableState
    dims: FacetDimsConfig
    cellSize: FacetCellSize
    padding: int
    selectedCell: typing.Optional[FacetCell] = dataclasses.field(
        default_factory=lambda: None
    )


@weave.type()
class Facet(panel.Panel):
    id = "Facet"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[FacetConfig] = None

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            table = table_state.TableState(self.input_node)
            self.config = FacetConfig(
                table=table,
                dims=FacetDimsConfig(
                    x=table.add_column(lambda row: graph.VoidNode()),
                    y=table.add_column(lambda row: graph.VoidNode()),
                    select=table.add_column(lambda row: graph.VoidNode()),
                    detail=table.add_column(lambda row: graph.VoidNode()),
                ),
                cellSize=FacetCellSize(w=50, h=50),
                padding=0,
                selectedCell=None,
            )
            self.set_x(options["x"])
            self.set_y(options["y"])
            self.config.table.enable_groupby(self.config.dims.x)
            self.config.table.enable_groupby(self.config.dims.y)
            if "select" in options:
                self.set_select(options["select"])
            if "detail" in options:
                self.set_detail(options["detail"])
            if "cell_size" in options:
                self.set_cell_size(*options["cell_size"])

    def debug_dim_select_functions(self):
        for dim in ["x", "y", "select", "detail"]:
            print(
                dim,
                self.config.table.columnSelectFunctions[
                    getattr(self.config.dims, dim)
                ].__repr__(),
            )

    def set_x(self, expr):
        self.config.table.update_col(self.config.dims.x, expr)

    def set_y(self, expr):
        self.config.table.update_col(self.config.dims.y, expr)

    def set_cell_size(self, w, h):
        self.config.cellSize = FacetCellSize(w=w, h=h)

    def set_select(self, expr):
        self.config.table.update_col(self.config.dims.select, expr)

    def set_detail(self, expr):
        self.config.table.update_col(self.config.dims.detail, expr)

    @weave.op(output_type=lambda input_type: input_type["self"].input_node.output_type)
    def selected(self):
        if self.config.selectedCell == None:
            from ..ops_arrow import list_

            return list_.make_vec_none(0)
        x_fn = self.config.table.columnSelectFunctions[self.config.dims.x]
        y_fn = self.config.table.columnSelectFunctions[self.config.dims.y]
        filtered = weave.ops.List.filter(
            self.input_node,
            lambda item: weave.ops.Boolean.bool_and(
                weave.ops.String.__eq__(
                    x_fn,
                    self.config.selectedCell["x"],
                ),
                weave.ops.String.__eq__(
                    y_fn,
                    self.config.selectedCell["y"],
                ),
            ),
        )
        return weave_internal.use(filtered)



import dataclasses
import json
import logging
import typing
import pyarrow as pa


from .context import get_error_on_non_vectorized_history_transform
from ...compile_domain import wb_gql_op_plugin
from ...api import op
from ... import weave_types as types
from .. import trace_tree, wb_domain_types as wdt
from ... import artifact_mem
from .. import wb_util
from ...ops_arrow import ArrowWeaveList, ArrowWeaveListType, convert
from ...op_def import map_type
from ... import engine_trace
from ... import errors
from ...wandb_interface import wandb_stream_table
from . import history_op_common
from ... import artifact_base, io_service
from .. import wbmedia
from ...ops_arrow.list_ import (
    PathItemType,
    PathType,
    weave_arrow_type_check,
)

tracer = engine_trace.tracer()


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history3_type(run: wdt.Run) -> types.Type:
    # TODO: Consider merging `_unflatten_history_object_type` into the main path
    return ArrowWeaveListType(
        _unflatten_history_object_type(history_op_common.refine_history_type(run))
    )


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(lambda inputs, inner: "historyKeys"),
    hidden=True,
)
def refine_history3_with_columns_type(
    run: wdt.Run, history_cols: list[str]
) -> types.Type:
    # TODO: Consider merging `_unflatten_history_object_type` into the main path
    return ArrowWeaveListType(
        _unflatten_history_object_type(
            history_op_common.refine_history_type(
                run,
                columns=history_op_common.get_full_columns_prefixed(run, history_cols),
            )
        )
    )


@op(
    name="run-history3_with_columns",
    refine_output_type=refine_history3_with_columns_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history3_with_columns(run: wdt.Run, history_cols: list[str]):
    return _get_history3(
        run, history_op_common.get_full_columns_prefixed(run, history_cols)
    )


@op(
    name="run-history3",
    refine_output_type=refine_history3_type,
    plugins=wb_gql_op_plugin(history_op_common.make_run_history_gql_field),
    output_type=ArrowWeaveListType(types.TypedDict({})),
    hidden=True,
)
def history3(run: wdt.Run):
    # TODO: This is now equivalent to hist2
    return history_op_common.mock_history_rows(run)


@dataclasses.dataclass
class PathTree:
    children: typing.Dict[PathItemType, "PathTree"] = dataclasses.field(
        default_factory=dict
    )
    data: typing.Optional[typing.Any] = None


def _get_history3(run: wdt.Run, columns=None):
    # 1. Get the flattened Weave-Type given HistoryKeys
    # 2. Read in the live set
    # 3. Raw-load each parquet file
    # 4. For each flattened column:
    #   a. If it contains a type requiring in-mem processing, then
    #       i. For liveset -> we must pass each item `_process_run_dict_item`, then use convert.to_arrow
    #       ii. For parquet -> we must extract the pyobj, us `_process_run_dict_item`, convert.to_arrow
    #   b. If the nested type contains a custom type or object type, then we must adjust the structure:
    #       i. For liveset -> we can pass each item to the simplified `_reconstruct_run_item`, then use convert.to_arrow(already_mapped = true)
    #       ii. For parquet -> we need to use the `_correct_awl` to adjust the structure.
    #   c. For all other types
    #       i. For liveset -> we can just use convert.to_arrow
    #       ii. For parquet -> no op
    # 5. Now we concat the converted liveset and parquet files
    # 6. Finally, unflatten the columns and return the result
    # 7. Optionally: verify the AWL

    artifact = artifact_mem.MemArtifact()

    # 1. Get the flattened Weave-Type given HistoryKeys
    flattened_object_type = history_op_common.refine_history_type(run, columns=columns)
    _verify_supported_types(flattened_object_type)
    final_type = _unflatten_history_object_type(flattened_object_type)

    # 2. Read in the live set
    raw_live_data = _get_live_data_from_run(run, columns=columns)

    # 3.a: Raw-load each parquet file
    raw_history_awl_tables = _read_raw_history_awl_tables(
        run, columns=columns, artifact=artifact
    )

    # 3.b: Collapse unions
    union_collapsed_awl_tables = [
        _collapse_unions(awl) for awl in raw_history_awl_tables
    ]

    # 4. Process all the data
    raw_history_pa_tables = [
        history_op_common.awl_to_pa_table(awl) for awl in union_collapsed_awl_tables
    ]

    run_path = wb_util.RunPath(
        run.gql["project"]["entity"]["name"],
        run.gql["project"]["name"],
        run.gql["name"],
    )
    (
        live_columns,
        live_columns_already_mapped,
        processed_history_pa_tables,
    ) = _process_all_columns(
        flattened_object_type, raw_live_data, raw_history_pa_tables, run_path, artifact
    )

    live_data_awl = _construct_live_data_awl(
        live_columns,
        live_columns_already_mapped,
        flattened_object_type,
        len(raw_live_data),
        artifact,
    )

    # 5.a Now we concat the converted liveset and parquet files
    concatted_awl = history_op_common.concat_awls(
        [
            live_data_awl,
            *{
                ArrowWeaveList(
                    table, object_type=flattened_object_type, artifact=artifact
                )
                for table in processed_history_pa_tables
            },
        ]
    )

    if len(concatted_awl) == 0:
        return convert.to_arrow([], types.List(final_type), artifact=artifact)

    sorted_table = history_op_common.sort_history_pa_table(
        history_op_common.awl_to_pa_table(concatted_awl)
    )

    # 6. Finally, unflatten the columns
    final_array = _unflatten_pa_table(sorted_table)

    # 7. Optionally: verify the AWL
    reason = weave_arrow_type_check(final_type, final_array)

    if reason != None:
        raise errors.WeaveWBHistoryTranslationError(
            f"Failed to effectively convert column of Gorilla Parquet History to expected history type: {reason}"
        )
    return ArrowWeaveList(
        final_array,
        final_type,
        artifact=artifact,
    )


def _construct_live_data_awl(
    live_columns: dict[str, list],
    live_columns_already_mapped: dict[str, list],
    flattened_object_type: types.TypedDict,
    num_rows: int,
    artifact: artifact_mem.MemArtifact,
) -> ArrowWeaveList:
    live_columns_to_data = [
        {k: v[i] for k, v in live_columns.items()} for i in range(num_rows)
    ]

    live_columns_already_mapped_to_data = [
        {k: v[i] for k, v in live_columns_already_mapped.items()}
        for i in range(num_rows)
    ]

    partial_live_data_awl = convert.to_arrow(
        live_columns_to_data,
        types.List(
            types.TypedDict(
                {
                    k: v
                    for k, v in flattened_object_type.property_types.items()
                    if k in live_columns
                }
            )
        ),
        artifact=artifact,
    )
    partial_live_data_already_mapped_awl = convert.to_arrow_from_list_and_artifact(
        live_columns_already_mapped_to_data,
        types.TypedDict(
            {
                k: v
                for k, v in flattened_object_type.property_types.items()
                if k in live_columns_already_mapped
            }
        ),
        artifact=artifact,
        py_objs_already_mapped=True,
    )

    field_names = []
    field_arrays = []
    for field_ndx, field in enumerate(partial_live_data_awl._arrow_data.type):
        field_names.append(field.name)
        field_arrays.append(partial_live_data_awl._arrow_data.field(field_ndx))

    for field_ndx, field in enumerate(
        partial_live_data_already_mapped_awl._arrow_data.type
    ):
        field_names.append(field.name)
        field_arrays.append(
            partial_live_data_already_mapped_awl._arrow_data.field(field_ndx)
        )

    return ArrowWeaveList(
        pa.StructArray.from_arrays(field_arrays, field_names),
        flattened_object_type,
        artifact=artifact,
    )


def _process_all_columns(
    flattened_object_type: types.TypedDict,
    raw_live_data: list[dict],
    raw_history_pa_tables: list[pa.Table],
    run_path: wb_util.RunPath,
    artifact: artifact_mem.MemArtifact,
):
    live_columns = {}
    live_columns_already_mapped = {}
    processed_history_pa_tables = [*raw_history_pa_tables]

    for col_name, col_type in flattened_object_type.property_types.items():
        raw_live_column = _extract_column_from_live_data(raw_live_data, col_name)
        processed_live_column = None
        if _column_type_requires_in_memory_transformation(col_type):
            _non_vectorized_warning(
                f"Encountered a history column requiring non-vectorized, in-memory processing: {col_name}: {col_type}"
            )
            processed_live_column = _process_column_in_memory(raw_live_column, run_path)
            live_columns[col_name] = processed_live_column

            for table_ndx, table in enumerate(processed_history_pa_tables):
                fields = [field.name for field in table.schema]
                new_col = _process_history_column_in_memory(
                    table[col_name], col_type, run_path, artifact
                )
                processed_history_pa_tables[table_ndx] = table.set_column(
                    fields.index(col_name), col_name, new_col
                )

        elif _is_directly_convertible_type(col_type):
            live_columns_already_mapped[col_name] = raw_live_column
            # Important that this runs before the else branch
            continue
        else:
            processed_live_column = _reconstruct_original_live_data_col(raw_live_column)
            live_columns_already_mapped[col_name] = processed_live_column

            for table_ndx, table in enumerate(processed_history_pa_tables):
                fields = [field.name for field in table.schema]
                new_col = _drop_types_from_encoded_types(
                    ArrowWeaveList(
                        table[col_name],
                        None,
                        artifact=artifact,
                    )
                )
                arrow_col = new_col._arrow_data
                if types.Timestamp().assign_type(types.non_none(col_type)):
                    arrow_col = arrow_col.cast("int64").cast(
                        pa.timestamp("ms", tz="UTC")
                    )
                processed_history_pa_tables[table_ndx] = table.set_column(
                    fields.index(col_name), col_name, arrow_col
                )

    return live_columns, live_columns_already_mapped, processed_history_pa_tables


unsupported_types = (
    wbmedia.ImageArtifactFileRefType,  # type: ignore
    wbmedia.AudioArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.BokehArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.VideoArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.Object3DArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.MoleculeArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.HtmlArtifactFileRef.WeaveType,  # type: ignore
    wbmedia.LegacyTableNDArrayType,
    wb_util.WbHistogram,
)


def _verify_supported_type_mapper(type_: types.Type) -> types.Type:
    from .. import artifact_fs

    type_ = types.non_none(type_)
    if isinstance(type_, unsupported_types + (artifact_fs.FilesystemArtifactFileType,)):
        raise errors.WeaveWBHistoryTranslationError(
            f"Unsupported type in history3: {type_}"
        )
    return type_


def _verify_supported_types(type_: types.Type) -> None:
    map_type(type_, _verify_supported_type_mapper)


def _non_vectorized_warning(message: str):
    logging.warning(message)
    if get_error_on_non_vectorized_history_transform():
        raise errors.WeaveWBHistoryTranslationError(message)


def _build_array_from_tree(tree: PathTree) -> pa.Array:
    children_struct_array = []
    if len(tree.children) > 0:
        mask = None
        children_data = {}
        for k, v in tree.children.items():
            children_data[k] = _build_array_from_tree(v)
            child_is_null = _is_null(children_data[k])
            if mask is None:
                mask = child_is_null
            else:
                mask = pa.compute.or_(mask, child_is_null)

        children_struct = pa.StructArray.from_arrays(
            children_data.values(), children_data.keys(), mask=mask
        )
        children_struct_array = [children_struct]
    children_arrays = [*(tree.data or []), *children_struct_array]

    if len(children_arrays) == 0:
        raise errors.WeaveWBHistoryTranslationError(
            "Cannot build array from empty tree"
        )
    elif len(children_arrays) == 1:
        return children_arrays[0]

    num_rows = len(children_arrays[0])

    return _union_from_column_data(num_rows, children_arrays)


def _unflatten_pa_table(array: pa.Table) -> pa.StructArray:
    cols = array.schema.names
    new_tree = PathTree(data=[])
    for col_name in cols:
        path = col_name.split(".")
        target = new_tree
        for part in path:
            if part not in target.children:
                target.children[part] = PathTree(data=[])
            target = target.children[part]
        if not isinstance(target.data, list):
            raise errors.WeaveWBHistoryTranslationError(
                f"Encountered unexpected data in PathTree: {type(target.data)}"
            )
        target.data.append(array[col_name].combine_chunks())

    # Recursively resolve the tree
    return _build_array_from_tree(new_tree)


def _union_collapse_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    object_type = types.non_none(awl.object_type)
    if types.TypedDict({}).assign_type(object_type):
        object_type = typing.cast(types.TypedDict, object_type)
        columns = list(object_type.property_types.keys())
        arrow_data = awl._arrow_data
        arrow_columns = [c.name for c in arrow_data.type]
        if all([c.startswith("_type_") for c in columns]):
            logging.warning(
                f"Encountered history store union: this requires a non-vectorized transformation on the order of number of rows (but fortunately does not require transforming to python). Path: {path}, Type: {object_type}"
            )
            num_rows = len(arrow_data)
            weave_type_members = []
            for col_name in columns:
                weave_type_members.append(object_type.property_types[col_name])

            union_arr = _union_from_column_data(
                num_rows,
                [
                    arrow_data.field(arrow_columns.index(col_name))
                    for col_name in columns
                ],
            )
            return ArrowWeaveList(
                union_arr, types.union(*weave_type_members), awl._artifact
            )

    return None


def _union_from_column_data(num_rows: int, columns: list[pa.Array]) -> pa.Array:
    type_array = pa.nulls(num_rows, type=pa.int8())
    offset_array = pa.nulls(num_rows, type=pa.int32())
    arrays = []
    for col_ndx, column_data in enumerate(columns):
        is_usable = pa.compute.invert(_is_null(column_data))
        new_data = column_data.filter(is_usable)
        arrays.append(new_data)
        type_array = pa.compute.replace_with_mask(
            type_array,
            is_usable,
            pa.array([col_ndx] * len(is_usable), type=pa.int8()),
        )
        offset_array = pa.compute.replace_with_mask(
            offset_array,
            is_usable,
            pa.array(list(range(len(new_data))), type=pa.int32()),
        )

    arrays.append(pa.nulls(1, type=pa.int8()))
    type_array = pa.compute.fill_null(type_array, len(arrays) - 1)
    offset_array = pa.compute.fill_null(offset_array, 0)

    return pa.UnionArray.from_dense(type_array, offset_array, arrays)


def _is_null(array: pa.Array) -> pa.Array:
    if isinstance(array, pa.ChunkedArray):
        array = array.combine_chunks()

    base_truth = pa.compute.is_null(array)

    if pa.types.is_struct(array.type):
        children_nulls = [_is_null(array.field(i)) for i in range(len(array.type))]
        if len(children_nulls) == 0:
            return pa.array([True] * len(array), type=pa.bool_())
        else:
            curr = children_nulls[0]
            for child in children_nulls[1:]:
                curr = pa.compute.and_(curr, child)
            return pa.compute.or_(curr, base_truth)

    return base_truth


def _collapse_unions(awl: ArrowWeaveList) -> ArrowWeaveList:
    return awl.map_column(_union_collapse_mapper)


def _drop_types_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    object_type = types.non_none(awl.object_type)
    if types.TypedDict(
        {"_type": types.optional(types.String()), "_val": types.Any()}
    ).assign_type(object_type):
        return awl.column("_val")
    return None


def _drop_types_from_encoded_types(awl: ArrowWeaveList) -> ArrowWeaveList:
    return awl.map_column(_drop_types_mapper)


def _get_live_data_from_run(run: wdt.Run, columns=None):
    raw_live_data = run.gql["sampledParquetHistory"]["liveData"]
    if columns is None:
        return raw_live_data
    column_set = set(columns)
    return [{k: v for k, v in row.items() if k in column_set} for row in raw_live_data]


def _extract_column_from_live_data(live_data: list[dict], column_name: str):
    return [row.get(column_name, None) for row in live_data]


def _process_column_in_memory(live_column: list, run_path: wb_util.RunPath) -> list:
    return [_process_live_object_in_memory(item, run_path) for item in live_column]


def _process_live_object_in_memory(
    live_data: typing.Any, run_path: wb_util.RunPath
) -> typing.Any:
    if isinstance(live_data, list):
        return [_process_live_object_in_memory(cell, run_path) for cell in live_data]
    if isinstance(live_data, dict):
        if live_data.get("_type") != None:
            return wb_util._process_run_dict_item(live_data, run_path)
        else:
            return {
                key: _process_live_object_in_memory(val, run_path)
                for key, val in live_data.items()
            }
    return live_data


def _process_history_column_in_memory(
    history_column: pa.Array,
    col_type: types.Type,
    run_path: wb_util.RunPath,
    artifact: artifact_base.Artifact,
) -> pa.Array:
    pyobj = history_column.to_pylist()
    processed_data = _process_column_in_memory(pyobj, run_path)
    awl = convert.to_arrow(
        processed_data,
        types.List(col_type),
        artifact=artifact,
    )
    return pa.chunked_array([awl._arrow_data])


# Copy from common - need to merge back
def _read_raw_history_awl_tables(
    run: wdt.Run,
    columns=None,
    artifact: typing.Optional[artifact_base.Artifact] = None,
) -> list[ArrowWeaveList]:
    io = io_service.get_sync_client()
    tables = []
    for url in run.gql["sampledParquetHistory"]["parquetUrls"]:
        local_path = io.ensure_file_downloaded(url)
        if local_path is not None:
            path = io.fs.path(local_path)
            awl = history_op_common.awl_from_local_parquet_path(
                path, None, columns=columns, artifact=artifact
            )
            awl.map_column(_parse_bytes_mapper)
            tables.append(awl)
    return tables

    # return history_op_common.process_history_awl_tables(tables)


def _parse_bytes_to_json(bytes: bytes):
    return json.loads(bytes.decode("utf-8"))


def _parse_bytes_mapper(
    awl: ArrowWeaveList, path: PathType
) -> typing.Optional[ArrowWeaveList]:
    obj_type = types.non_none(awl.object_type)
    if types.Bytes().assign_type(obj_type):
        _non_vectorized_warning(
            f"Encountered bytes in column {path}, requires in-memory processing"
        )
        data = [
            _parse_bytes_to_json(row) if row is not None else None
            for row in awl._arrow_data.to_pylist()
        ]
        wb_type = types.TypeRegistry.type_of(data)
        if not hasattr(wb_type, "object_type"):
            raise errors.WeaveWBHistoryTranslationError(
                f"Expected type with object_type attribute, got {wb_type}"
            )
        wb_type = typing.cast(types.List, wb_type)
        return convert.to_arrow_from_list_and_artifact(
            data,
            wb_type.object_type,
            artifact=awl._artifact or artifact_mem.MemArtifact(),
            py_objs_already_mapped=True,
        )
    return None


def _column_type_requires_in_memory_transformation(col_type: types.Type):
    return _column_type_requires_in_memory_transformation_recursive(col_type)


def _is_directly_convertible_type(col_type: types.Type):
    return _is_directly_convertible_type_recursive(col_type)


_weave_types_requiring_in_memory_transformation = (
    # wbmedia.ImageArtifactFileRefType,  # type: ignore
    # wbmedia.AudioArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.BokehArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.VideoArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.Object3DArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.MoleculeArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.HtmlArtifactFileRef.WeaveType,  # type: ignore
    # wbmedia.LegacyTableNDArrayType,
    # wb_util.WbHistogram,
    trace_tree.WBTraceTree.WeaveType,  # type: ignore
    types.UnknownType,
)


def _column_type_requires_in_memory_transformation_recursive(col_type: types.Type):
    if types.NoneType().assign_type(col_type):
        return False
    non_none_type = types.non_none(col_type)
    if isinstance(non_none_type, types.UnionType):
        return True
    elif isinstance(non_none_type, _weave_types_requiring_in_memory_transformation):
        return True
    elif isinstance(non_none_type, types.List):
        return _column_type_requires_in_memory_transformation_recursive(
            non_none_type.object_type
        )
    elif isinstance(non_none_type, types.TypedDict):
        for k, v in non_none_type.property_types.items():
            if _column_type_requires_in_memory_transformation_recursive(v):
                return True
        return False
    else:
        return False


def _is_directly_convertible_type_recursive(col_type: types.Type):
    if types.NoneType().assign_type(col_type):
        return True
    non_none_type = types.non_none(col_type)
    if types.union(
        *[
            types.Number(),
            types.Int(),
            types.Float(),
            types.String(),
            types.Boolean(),
        ]
    ).assign_type(col_type):
        return True
    elif isinstance(non_none_type, types.List):
        return _is_directly_convertible_type_recursive(non_none_type.object_type)
    elif isinstance(non_none_type, types.TypedDict):
        for k, v in non_none_type.property_types.items():
            if not _is_directly_convertible_type_recursive(v):
                return False
        return True
    else:
        return False


def _reconstruct_original_live_data_col(row: list):
    return [_reconstruct_original_live_data_cell(cell) for cell in row]


def _reconstruct_original_live_data_cell(live_data: typing.Any) -> typing.Any:
    if isinstance(live_data, list):
        return [_reconstruct_original_live_data_cell(cell) for cell in live_data]
    if isinstance(live_data, dict):
        if wandb_stream_table.is_weave_encoded_history_cell(live_data):
            return live_data["_val"]
        return {
            key: _reconstruct_original_live_data_cell(val)
            for key, val in live_data.items()
        }
    return live_data


def _unflatten_history_object_type(obj_type: types.TypedDict) -> types.TypedDict:
    # Need to combine top-level maps. Note: this is not a recursive function
    # since the flattening only happens top level in gorilla.
    dict_keys: dict[str, types.Type] = {}
    for key, val in obj_type.property_types.items():
        path_parts = key.split(".")
        prop_types = dict_keys

        for part in path_parts[:-1]:
            found = False
            prop_types_inner: dict[str, types.Type] = {}
            default_dict = types.TypedDict(prop_types_inner)
            if part not in prop_types:
                prop_types[part] = default_dict
                prop_types = prop_types_inner
            elif part in prop_types:
                prop_type = prop_types[part]
                if isinstance(prop_type, types.TypedDict):
                    prop_types = prop_type.property_types
                    found = True
                elif isinstance(prop_type, types.UnionType):
                    for member in prop_type.members:
                        if isinstance(member, types.TypedDict):
                            prop_types = member.property_types
                            found = True
                            break

                if not found:
                    prop_types[part] = types.union(prop_type, default_dict)
                    prop_types = prop_types_inner

        final_part = path_parts[-1]
        if final_part not in prop_types:
            prop_types[final_part] = val
        else:
            prop_types[final_part] = types.union(val, prop_types[final_part])

    return types.TypedDict(dict_keys)



import os
import xgboost
import typing

import weave


class XGBoostModelType(weave.types.Type):
    name = "xgboost_model"
    instance_class = xgboost.core.Booster
    instance_classes = xgboost.core.Booster

    def save_instance(cls, obj, artifact, name):
        os.makedirs(artifact._write_dirname, exist_ok=True)
        f = os.path.join(artifact._write_dirname, f"{name}.json")
        obj.save_model(f)

    def load_instance(self, artifact, name, extra=None):
        f = os.path.join(artifact._read_dirname, f"{name}.json")
        model_xgb = xgboost.Booster()
        model_xgb.load_model(f)
        return model_xgb


@weave.weave_class(weave_type=XGBoostModelType)
class XGBoostModelOps:
    @weave.op()
    def predict(self, data: typing.Any) -> typing.Any:
        return self.predict(xgboost.DMatrix(data))


class XGBoostHyperparams(typing.TypedDict):
    learning_rate: float


@weave.op()
def xgboost_train(
    xy: typing.Any, hyperparams: XGBoostHyperparams
) -> xgboost.core.Booster:
    return xgboost.train(
        hyperparams, xgboost.DMatrix(xy["X"], label=xy["y"].to_numpy()), 100
    )



import dataclasses
import datetime
import typing
import inspect
import functools
import json
from collections.abc import Iterable


from . import box
from . import errors
from . import mappers_python
from . import timestamp as weave_timestamp


if typing.TYPE_CHECKING:
    from .artifact_fs import FilesystemArtifact
    from . import artifact_base


def to_weavejs_typekey(k: str) -> str:
    if k == "object_type":
        return "objectType"
    elif k == "property_types":
        return "propertyTypes"
    return k


def js_to_py_typename(typename: str) -> str:
    # WeaveJS overrides the table column type if the column name is id-like. See
    # use of `SPECIAL_ID_COLUMN_NAMES` in `mediaTypes.ts`. Unlike other types,
    # the "id" type does not describe the underlying data type. Furthermore,
    # there are no ops based on the "id" type. For that reason, I opted to not
    # put the id type in the type system for now, and just hap it to string.
    if typename == "id":
        return "string"
    return typename


def all_subclasses(cls):
    # Note using a list here... this doesn't dedupe star pattern!
    # But it does preserve tree order.
    return list(cls.__subclasses__()) + [
        s for c in cls.__subclasses__() for s in all_subclasses(c)
    ]


def get_type_classes():
    return all_subclasses(Type)


def instance_class_to_potential_type_map():
    res = {}
    type_classes = get_type_classes()
    for type_class in type_classes:
        for instance_class in type_class._instance_classes():
            res.setdefault(instance_class, []).append(type_class)

    # These are errors and we should fix them
    # TODO
    # for ic, tcs in res.items():
    #     if len(tcs) > 1:
    #         print("LONG TC", ic, tcs)
    return res


@functools.cache
def instance_class_to_potential_type(cls):
    mapping = instance_class_to_potential_type_map()
    exact_type_classes = mapping.get(cls)
    if exact_type_classes is not None:
        return exact_type_classes
    result = []
    for instance_class, type_classes in mapping.items():
        if issubclass(cls, instance_class):
            result += type_classes
    return result


@functools.cache
def type_name_to_type_map():
    res = {}
    type_classes = get_type_classes()
    for type_class in type_classes:
        name = type_class.name
        if name is not None:
            res[name] = type_class
    return res


@functools.cache
def type_name_to_type(type_name):
    mapping = type_name_to_type_map()
    return mapping.get(js_to_py_typename(type_name))


class TypeRegistry:
    @staticmethod
    def has_type(obj):
        return bool(instance_class_to_potential_type(type(obj)))

    @staticmethod
    def type_class_of(obj):
        type_classes = instance_class_to_potential_type(type(obj))
        if not type_classes:
            # return UnknownType()
            raise errors.WeaveTypeError("no Type for obj: %s" % obj)
        return type_classes[-1]

    @staticmethod
    def type_of(obj: typing.Any) -> "Type":
        obj_type = type_name_to_type("tagged").type_of(obj)
        if obj_type is not None:
            return obj_type

        # use reversed instance_class_to_potential_type so our result
        # is the most specific type.

        # The type of graph.Node objects is Function, but we also mixin
        # NodeMethodsClass with Node, which may also have a Weave type.
        # Therefore we have multiple valid types. But we need type_of() to
        # be function in that case. Ie when a Node is mixed in, its lazy
        # representation of whatever the type is, not the actual type.
        #
        # TODO: Its a small that we need to hardcode here. Fix this.
        potential_types = instance_class_to_potential_type(type(obj))  # type: ignore
        if "function" in (t.name for t in potential_types):
            potential_types = [Function]

        for type_ in reversed(potential_types):
            obj_type = type_.type_of(obj)
            if obj_type is not None:
                return obj_type
        # return UnknownType()
        raise errors.WeaveTypeError("no Type for obj: (%s) %s" % (type(obj), obj))

    @staticmethod
    def type_from_dict(d: typing.Union[str, dict]) -> "Type":
        # The javascript code sends simple types as just strings
        # instead of {'type': 'string'} for example
        type_name = d["type"] if isinstance(d, dict) else d
        type_ = type_name_to_type(type_name)
        if type_ is None:
            raise errors.WeaveSerializeError("Can't deserialize type from: %s" % d)
        return type_.from_dict(d)


def _clear_global_type_class_cache():
    instance_class_to_potential_type.cache_clear()
    type_name_to_type_map.cache_clear()
    type_name_to_type.cache_clear()


def _cached_hash(self):
    try:
        return self.__dict__["_hash"]
    except KeyError:
        hashed = hash((self.__class__, self._hashable()))
        self.__dict__["_hash"] = hashed
        return hashed


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


# Addapted from https://stackoverflow.com/questions/18126552/how-to-run-code-when-a-class-is-subclassed
class _TypeSubclassWatcher(type):
    def __init__(cls, name, bases, clsdict):
        # This code will run whenever `Type` is subclassed!
        # Bust the cache!
        _clear_global_type_class_cache()
        super(_TypeSubclassWatcher, cls).__init__(name, bases, clsdict)

        # Set _base_type to the first base class
        if bases and bases[0] != Type:
            if cls.__dict__.get("_base_type") is None:
                cls._base_type = bases[0]

        # Override the dataclass __hash__ with our own version
        cls.__hash__ = _cached_hash


@dataclasses.dataclass(frozen=True)
class Type(metaclass=_TypeSubclassWatcher):
    # This is like Typescript "extends", and is populated by default by
    # Python inheritance in our metaclass.
    _base_type: typing.ClassVar[typing.Optional[typing.Type["Type"]]] = None

    instance_class: typing.ClassVar[typing.Optional[type]]
    instance_classes: typing.ClassVar[
        typing.Union[type, typing.List[type], None]
    ] = None

    _type_attrs = None
    _hash = None

    def _hashable(self):
        return tuple((k, hash(t)) for k, t in self.type_vars_tuple)

    def __lt__(self, other):
        return hash(self) < hash(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __name__(self) -> str:
        return self.__repr__()

    def simple_str(self) -> str:
        return str(self)

    @classmethod
    def _instance_classes(cls):
        """Helper to get instance_classes as iterable."""
        if cls.instance_classes is None:
            return ()
        if isinstance(cls.instance_classes, Iterable):
            return cls.instance_classes
        return (cls.instance_classes,)

    @classmethod
    def type_attrs(cls):
        type_attrs = []
        for field in dataclasses.fields(cls):
            if (inspect.isclass(field.type) and issubclass(field.type, Type)) or (
                field.type.__origin__ == typing.Union
                and any(issubclass(a, Type) for a in field.type.__args__)
            ):
                type_attrs.append(field.name)
        return type_attrs

    @property
    def type_vars_tuple(self):
        type_vars = []
        for field in self.type_attrs():
            type_vars.append((field, getattr(self, field)))
        return tuple(type_vars)

    @property
    def type_vars(self) -> dict[str, "Type"]:
        return dict(self.type_vars_tuple)

    @classmethod
    def is_instance(cls, obj):
        for ic in cls._instance_classes():
            if isinstance(obj, ic):
                return ic
        return None

    @classmethod
    def type_of(cls, obj) -> typing.Optional["Type"]:
        if not cls.is_instance(obj):
            return None
        return cls.type_of_instance(obj)

    @classmethod
    def type_of_instance(cls, obj):
        # Default implementation for Types that take no arguments.
        return cls()

    @classproperty
    def name(cls):
        if cls == Type:
            return "type"
        return cls.__name__.removesuffix("Type")

    @property  # type: ignore
    def instance_class(self):
        return self._instance_classes()[-1]

    def assign_type(self, next_type: "Type") -> bool:
        # assign_type needs to be as fast as possible, so there are optimizations
        # throughout this code path, like checking for class equality instead of using isinstance

        # Unions or union-like things (tagged unions) must be handled first
        if next_type.name == "tagged":
            next_type = next_type._assignment_form  # type: ignore

        is_assignable_to = next_type._is_assignable_to(self)
        if is_assignable_to:
            return True

        # Check class attribute directly instead of isinstance for performance
        if next_type.__class__ == UnionType:
            return all(self.assign_type(t) for t in next_type.members)  # type: ignore
        if self.__class__ == UnionType:
            return any(t.assign_type(next_type) for t in self.members)  # type: ignore

        return self._assign_type_inner(next_type)

    def _is_assignable_to(self, other_type) -> typing.Optional[bool]:
        return None

    def _assign_type_inner(self, next_type: "Type") -> bool:
        # First check that we match next_type class or one of
        # its bases.
        next_type_class = next_type.__class__
        while True:
            if self.__class__ == next_type_class:
                break
            elif next_type_class._base_type is None:
                # nothing left in base chain, no match
                return False
            next_type_class = next_type_class._base_type

        # Now check that all type vars match
        for prop_name in self.type_vars:
            if not hasattr(next_type, prop_name):
                return False
            if not getattr(self, prop_name).assign_type(getattr(next_type, prop_name)):
                return False
        return True

    @classmethod
    def class_to_dict(cls) -> dict[str, typing.Any]:
        d = {"type": cls.name}
        if cls._base_type is not None:
            d["_base_type"] = cls._base_type.class_to_dict()
        return d

    def to_dict(self) -> typing.Union[dict, str]:
        d = {"type": self.name}
        d.update(self._to_dict())
        return d

    def _to_dict(self) -> dict:
        fields = dataclasses.fields(self.__class__)
        type_props = {}
        if self._base_type is not None:
            type_props["_base_type"] = {"type": self._base_type.name}
        for field in fields:
            # TODO: I really don't like this change. Only needed because
            # FileType has optional fields... Remove?
            attr = getattr(self, field.name)
            if not attr:
                continue
            type_props[to_weavejs_typekey(field.name)] = attr.to_dict()
        return type_props

    @classmethod
    def from_dict(cls, d):
        fields = dataclasses.fields(cls)
        type_attrs = {}
        for field in fields:
            field_name = to_weavejs_typekey(field.name)
            if field_name in d:
                type_attrs[field.name] = TypeRegistry.type_from_dict(d[field_name])
        return cls(**type_attrs)

    def save_instance(
        self, obj, artifact, name
    ) -> typing.Optional[typing.Union[list[str], "artifact_base.ArtifactRef"]]:
        d = None
        try:
            d = self.instance_to_dict(obj)
        except NotImplementedError:
            pass
        if d is None:
            raise errors.WeaveSerializeError(
                "Object is not serializable. Provide instance_<to/from>_dict or <save/load>_instance methods on Type: %s"
                % self
            )
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(d, f)
        return None

    def load_instance(
        self,
        artifact: "FilesystemArtifact",
        name: str,
        extra: typing.Optional[list[str]] = None,
    ) -> typing.Any:
        with artifact.open(f"{name}.object.json") as f:
            d = json.load(f)
        return self.instance_from_dict(d)

    def instance_to_dict(self, obj):
        raise NotImplementedError

    def instance_from_dict(self, d):
        raise NotImplementedError

    @classmethod
    def make(cls, kwargs={}):
        return cls._make(cls, kwargs)

    @staticmethod
    def _make(cls, kwargs={}):
        raise Exception("Please import `weave` to use `Type.make`.")


# _PlainStringNamedType should only be used for backward compatibility with
# legacy WeaveJS code.
class _PlainStringNamedType(Type):
    def to_dict(self):
        # A basic type is serialized as just its string name.
        return self.name


class BasicType(_PlainStringNamedType):
    def save_instance(self, obj, artifact, name):
        if artifact is None:
            raise errors.WeaveSerializeError(
                "save_instance invalid when artifact is None for type: %s" % self
            )
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            return json.load(f)


class InvalidPy:
    pass


class Invalid(BasicType):
    name = "invalid"

    instance_classes = InvalidPy


class UnknownType(BasicType):
    name = "unknown"

    # TODO: we should strictly define the rules for unknown types. As is, we
    # allow anything to be assigned to unknown, but unknown can't be assigned to
    # anything else. This is effectively the same as Any. Is that correct?
    # Currently we have a single special case in List assign: `if
    # isinstance(next_type, List) and next_type.object_type == UnknownType():`
    #
    # I _think_ we want 'unknown' to be assigned to anything, and this would
    # allow us to remove the `_assign_type_inner` special case in List.
    #
    # This change would need to be implemented in the core `Type.assign_type`
    # method
    def _assign_type_inner(self, next_type):
        return True


# TODO: Tim has this as ConstType(None), but how is that serialized?


class NoneType(BasicType):
    name = "none"
    instance_classes = [type(None), box.BoxedNone]

    # If we're using NoneType in a place where we expect a list, the object_type
    # of that list is also NoneType, due to nullability.
    @property
    def object_type(self):
        return NoneType()

    def save_instance(self, obj, artifact, name):
        # BoxedNone is actually a box, not a subclass of bool, since
        # we can't subclass bool in Python. So we unbox it here.
        if isinstance(obj, box.BoxedNone):
            obj = obj.val
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(obj, f)


# TODO: use this seminal value all the time! We use it in is_optional()
#     to check if something is unioned with NoneType. But its too easy for
#     users to instantiate a different version of this. Maybe could muck
#     with Python's __new__ to fix that
none_type = NoneType()


class Any(BasicType):
    name = "any"

    def _assign_type_inner(self, next_type):
        return True


@dataclasses.dataclass(frozen=True)
class Const(Type):
    name = "const"
    val_type: Type
    val: typing.Any

    def __post_init__(self):
        if isinstance(self.val_type, UnionType):
            # Const Unions are invalid. If something is Const, that means we know what its value
            # is, which means it is definitely not a Union. Replace val_type with the actual type
            # here.
            self.__dict__["val_type"] = TypeRegistry.type_of(self.val)

    def _hashable(self):
        val_id = id(self.val)
        try:
            val_id = hash(self.val)
        except TypeError:
            pass
        return (hash(self.val_type), val_id)

    def __getattr__(self, attr):
        return getattr(self.val_type, attr)

    @classmethod
    def type_of_instance(cls, obj):
        return cls(TypeRegistry.type_of(obj.val), obj.val)

    def _is_assignable_to(self, other_type) -> typing.Optional[bool]:
        if other_type.__class__ != Const:
            return other_type.assign_type(self.val_type)
        return None

    def _assign_type_inner(self, next_type):
        if isinstance(next_type, Const):
            # This does a check on class equality, so won't work for
            # fancier types. We can fix later if we need to.
            if self.__class__ != next_type.__class__:
                return False
            if self.val == None or self.val == next_type.val:
                return True

        # TODO: should be "return False"! But we allow assign non const
        #     to const now. Because type_of("x") is String, not Const(String, "x").
        #     Why? Because we don't want to detect a column of non-uniform
        #     strings as a giant union.
        #     For example see test_plot_constants_assign
        return self.val_type.assign_type(next_type)

    @classmethod
    def from_dict(cls, d):
        w_type = TypeRegistry.type_from_dict(d["valType"])
        val = d["val"]
        # try:
        #     val = w_type.instance_from_dict(val)
        # except NotImplementedError:
        #     pass
        return cls(w_type, val)

    def _to_dict(self):
        val = self.val
        # try:
        #     val = self.val_type.instance_to_dict(val)
        # except NotImplementedError:
        #     pass
        return {"valType": self.val_type.to_dict(), "val": val}

    def __repr__(self):
        return "Const(%s, %s)" % (self.val_type, self.val)


class String(BasicType):
    name = "string"
    instance_classes = str

    # Just for String, we use a Const Type
    # TODO: this sucks! Maybe we need a const object?
    # but how does user code know to use it?
    # @classmethod
    # def type_of_instance(cls, obj):
    #     return Const(cls(), obj)


# TODO: support this in weave0. for weave1 use only now (dont send over wire)
class Bytes(BasicType):
    name = "bytes"
    instance_classes = bytes


class Number(BasicType):
    name = "number"


class Float(Number):
    instance_classes = float
    name = "float"

    def _assign_type_inner(self, other_type: Type):
        # Float if either side is a number
        return isinstance(other_type, Number)


class Int(Number):
    name = "int"
    instance_classes = int

    @classmethod
    def is_instance(cls, obj):
        # Special case, in Python bool isinstance of obj!
        if type(obj) == bool:
            return False
        return super().is_instance(obj)

    def _assign_type_inner(self, other_type: Type):
        # Become Float if rhs is Float
        return isinstance(other_type, Number) and not isinstance(other_type, Float)


class Boolean(BasicType):
    instance_classes = [bool, box.BoxedBool]
    name = "boolean"

    def save_instance(self, obj, artifact, name):
        # BoxedBool is actually a box, not a subclass of bool, since
        # we can't subclass bool in Python. So we unbox it here.
        if isinstance(obj, box.BoxedBool):
            obj = obj.val
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(obj, f)


class LegacyDate(Type):
    # TODO: For historic reasons we had "date" and "timestamp" types.  Now we only
    # use "timestamp" but we need to keep this around for backwards compatibility.
    name = "date"

    def save_instance(self, obj, artifact, name):
        if artifact is None:
            raise errors.WeaveSerializeError(
                "save_instance invalid when artifact is None for type: %s" % self
            )
        with artifact.new_file(f"{name}.object.json") as f:
            v = weave_timestamp.python_datetime_to_ms(obj)
            json.dump(v, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            v = json.load(f)
            return weave_timestamp.ms_to_python_datetime(v)


class TimeDelta(Type):
    name = "timedelta"
    instance_classes = datetime.timedelta

    def save_instance(self, obj: datetime.timedelta, artifact, name):
        if artifact is None:
            raise errors.WeaveSerializeError(
                "save_instance invalid when artifact is None for type: %s" % self
            )
        with artifact.new_file(f"{name}.object.json") as f:
            # Uses internal python representation of timedelta for serialization
            v = [obj.days, obj.seconds, obj.microseconds]
            json.dump(v, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            [days, seconds, microseconds] = json.load(f)
            return datetime.timedelta(days, seconds, microseconds)


class Timestamp(Type):
    name = "timestamp"
    instance_classes = datetime.datetime

    def from_isostring(self, iso: str) -> datetime.datetime:
        # NOTE: This assumes ISO 8601 format from GQL endpoints, it does NOT
        # support RFC 3339 strings with a "Z" at the end before python 3.11
        return datetime.datetime.fromisoformat(iso)

    def save_instance(self, obj, artifact, name):
        if artifact is None:
            raise errors.WeaveSerializeError(
                "save_instance invalid when artifact is None for type: %s" % self
            )
        with artifact.new_file(f"{name}.object.json") as f:
            v = weave_timestamp.python_datetime_to_ms(obj)
            json.dump(v, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            v = json.load(f)
            return weave_timestamp.ms_to_python_datetime(v)


@dataclasses.dataclass(frozen=True)
class UnionType(Type):
    name = "union"

    members: list[Type]

    def __init__(self, *members: Type) -> None:
        all_members = []
        for mem in members:
            if isinstance(mem, UnionType):
                for memmem in mem.members:
                    if memmem not in all_members:
                        all_members.append(memmem)
            else:
                if mem not in all_members:
                    all_members.append(mem)
        # # Remove UnknownType if there are other types
        # non_unknown_members = [
        #     mem for mem in all_members if not isinstance(mem, UnknownType)
        # ]
        # if non_unknown_members:
        #     all_members = non_unknown_members
        if not all_members:
            raise errors.WeaveInternalError("Attempted to construct empty union")
        if len(all_members) == 1:
            raise errors.WeaveInternalError(
                "Attempted to construct union with only one member, did you mean to use union()?"
            )
        object.__setattr__(self, "members", all_members)

    # def __repr__(self):
    #     return "UnionType(%s)" % ", ".join(repr(mem) for mem in self.members)

    def __eq__(self, other):
        if not isinstance(other, UnionType):
            return False
        # Order matters, it affects layout.
        # return self.members == other.members
        return set(self.members) == set(other.members)

    def _hashable(self):
        return tuple(hash(mem) for mem in sorted(self.members))

    def is_simple_nullable(self):
        return len(set(self.members)) == 2 and none_type in set(self.members)

    # def instance_to_py(self, obj):
    #     # Figure out which union member this obj is, and delegate to that
    #     # type.
    #     for member_type in self.members:
    #         if member_type.type_of(obj) is not None:
    #             return member_type.instance_to_py(obj)
    #     raise Exception('invalid')
    @classmethod
    def from_dict(cls, d):
        return merge_many_types(
            [TypeRegistry.type_from_dict(mem) for mem in d["members"]]
        )

    def _to_dict(self):
        return {"members": [mem.to_dict() for mem in self.members]}

    def __getattr__(self, attr):
        has_none = False
        results = []
        for mem in self.members:
            if mem == none_type:
                has_none = True
            else:
                results.append(getattr(mem, attr))
        if len(results) == 0:
            raise errors.WeaveTypeError(
                f"Attempt to get attribute {attr} from UnionType with no non-none members"
            )
        first_result = results[0]
        if isinstance(first_result, Type):
            if any((not isinstance(res, Type)) for res in results):
                raise errors.WeaveTypeError(
                    f"Attempt to get attribute {attr} from UnionType with inconsistent types"
                )
            if has_none:
                results.append(none_type)
            return union(*results)
        if isinstance(first_result, dict):
            if any((not isinstance(res, dict)) for res in results):
                raise errors.WeaveTypeError(
                    f"Attempt to get attribute {attr} from UnionType with inconsistent types"
                )
            all_keys = set(k for res in results for k in res.keys())
            if has_none:
                results.append({})
            return {
                k: union(*[res.get(k, NoneType()) for res in results]) for k in all_keys
            }
        raise errors.WeaveTypeError(
            f"Attempt to get attribute {attr} from UnionType resulted in type {type(first_result)}, expected Type or dict"
        )


@dataclasses.dataclass(frozen=True)
class List(Type):
    name = "list"
    instance_classes = [set, list]

    object_type: Type = dataclasses.field(default_factory=Any)

    @classmethod
    def type_of_instance(cls, obj):
        if not obj:
            return cls(UnknownType())
        list_obj_type = TypeRegistry.type_of(obj[0])
        for item in obj[1:]:
            obj_type = TypeRegistry.type_of(item)
            if obj_type is None:
                raise Exception("can't detect type for object: %s" % item)
            list_obj_type = merge_types(list_obj_type, obj_type)
        return cls(list_obj_type)

    def _assign_type_inner(self, next_type):
        if isinstance(next_type, List) and next_type.object_type == UnknownType():
            return True
        return super()._assign_type_inner(next_type)

    def save_instance(self, obj, artifact, name):
        serializer = mappers_python.map_to_python(self, artifact)
        result = serializer.apply(obj)
        with artifact.new_file(f"{name}.list.json") as f:
            json.dump(result, f, allow_nan=False)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.list.json") as f:
            result = json.load(f)
        mapper = mappers_python.map_from_python(self, artifact)
        mapped_result = mapper.apply(result)

        # TODO: scan through for ID
        from .ops_primitives.list_ import object_lookup

        if extra is not None:
            # return object_lookup.resolve_fn(mapped_result, extra[0])
            return mapped_result[int(extra[0])]
        return mapped_result


@dataclasses.dataclass(frozen=True)
class TypedDict(Type):
    name = "typedDict"
    instance_classes = [dict]
    property_types: dict[str, Type] = dataclasses.field(default_factory=dict)

    # See: https://peps.python.org/pep-0655/
    # in Typescript this is like key?: value
    not_required_keys: set[str] = dataclasses.field(default_factory=set)

    def _hashable(self):
        return tuple(
            (k, hash(v), k in self.not_required_keys)
            for k, v in self.property_types.items()
        )

    def _assign_type_inner(self, other_type):
        if isinstance(other_type, Dict):
            for ptype in self.property_types.values():
                if not ptype.assign_type(other_type.object_type):
                    return False
            return True

        if not isinstance(other_type, TypedDict):
            return False

        for k, ptype in self.property_types.items():
            if k in self.not_required_keys and k not in other_type.property_types:
                continue
            if k not in other_type.property_types or not ptype.assign_type(
                other_type.property_types[k]
            ):
                return False
        return True

    def _to_dict(self):
        property_types = {}
        for key, type_ in self.property_types.items():
            property_types[key] = type_.to_dict()
        result = {"propertyTypes": property_types}
        if self.not_required_keys:
            result["notRequiredKeys"] = list(self.not_required_keys)
        return result

    @classmethod
    def from_dict(cls, d):
        property_types = {}
        for key, type_ in d["propertyTypes"].items():
            property_types[key] = TypeRegistry.type_from_dict(type_)
        not_required_keys = set()
        if "notRequiredKeys" in d:
            not_required_keys = set(d["notRequiredKeys"])
        return cls(property_types, not_required_keys)

    @classmethod
    def type_of_instance(cls, obj):
        property_types = {}
        for k, v in obj.items():
            property_types[k] = TypeRegistry.type_of(v)
        return cls(property_types)

    def save_instance(self, obj, artifact, name):
        serializer = mappers_python.map_to_python(self, artifact)
        result = serializer.apply(obj)
        with artifact.new_file(f"{name}.typedDict.json") as f:
            json.dump(result, f, allow_nan=False)

    def load_instance(self, artifact, name, extra=None):
        # with artifact.open(f'{name}.type.json') as f:
        #     obj_type = TypeRegistry.type_from_dict(json.load(f))
        with artifact.open(f"{name}.typedDict.json") as f:
            result = json.load(f)
        mapper = mappers_python.map_from_python(self, artifact)
        mapped_result = mapper.apply(result)
        if extra is not None:
            return mapped_result[extra[0]]
        return mapped_result


@dataclasses.dataclass(frozen=True)
class Dict(Type):
    name = "dict"

    key_type: Type = String()
    object_type: Type = Any()

    @property
    def property_types(self):
        class DictPropertyTypes:
            def values(_self):
                return [self.object_type]

            def get(_self, _):
                return self.object_type

        return DictPropertyTypes()

    def __post_init__(self):
        # Note this differs from Python's Dict in that keys are always strings!
        # TODO: consider if we can / should accept key_type. Would make JS side
        # harder since js objects can only have string keys.
        if not String().assign_type(self.key_type):
            raise Exception("Dict only supports string keys!")

    def _assign_type_inner(self, other_type):
        if isinstance(other_type, TypedDict):
            return all(
                self.object_type.assign_type(t)
                for t in other_type.property_types.values()
            )
        return super()._assign_type_inner(other_type)

    @classmethod
    def type_of_instance(cls, obj):
        value_type = UnknownType()
        for k, v in obj.items():
            if not isinstance(k, str):
                raise Exception("Dict only supports string keys!")
            value_type = value_type.assign_type(TypeRegistry.type_of(v))
        return cls(String(), value_type)


@dataclasses.dataclass(frozen=True)
class ObjectType(Type):
    def __init__(self, **attr_types: Type):
        self.__dict__["attr_types"] = attr_types
        for k, v in attr_types.items():
            self.__dict__[k] = v

    @property
    def type_vars(self):
        if self.__class__ == ObjectType:
            return self.attr_types
        else:
            return super().type_vars

    def property_types(self) -> dict[str, Type]:
        return self.type_vars

    @classmethod
    def type_of_instance(cls, obj):
        variable_prop_types = {}
        for prop_name in cls.type_attrs():
            prop_type = TypeRegistry.type_of(getattr(obj, prop_name))
            # print("TYPE_OF", cls, prop_name, prop_type, type(getattr(obj, prop_name)))
            variable_prop_types[prop_name] = prop_type
        return cls(**variable_prop_types)

    def _to_dict(self) -> dict:
        d: dict = {"_is_object": True}
        d = self.class_to_dict()
        # TODO: we don't need _is_object, now that we have base_type everywhere.
        # Remove the check for self._base_type.__class__ != ObjectType, and get
        # rid of _is_object (need to update frontend as well).
        d["_is_object"] = True
        for k, prop_type in self.property_types().items():
            d[to_weavejs_typekey(k)] = prop_type.to_dict()
        return d

    def save_instance(self, obj, artifact, name):
        serializer = mappers_python.map_to_python(self, artifact)

        result = serializer.apply(obj)
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(result, f, allow_nan=False)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            result = json.load(f)
        mapper = mappers_python.map_from_python(self, artifact)
        return mapper.apply(result)


@dataclasses.dataclass(frozen=True)
class TypeType(ObjectType):
    name = "type"

    instance_classes = [Type]
    attr_types: dict[str, Type] = dataclasses.field(default_factory=dict)

    def property_types(self) -> dict[str, Type]:
        return self.attr_types

    def _hashable(self):
        return tuple((k, hash(v)) for k, v in self.property_types().items())

    @classmethod
    def type_of_instance(cls, obj):
        from . import infer_types

        attr_types = {}
        for field in dataclasses.fields(obj):
            attr_types[field.name] = infer_types.python_type_to_type(field.type)
        return cls(attr_types)

    def _to_dict(self):
        # we ensure we match the layout of ObjectType, so WeaveJS
        # can handle it the same way.
        d = {"_is_object": True}
        for k, t in self.attr_types.items():
            d[k] = t.to_dict()
        return d

    @classmethod
    def from_dict(cls, d):
        # weavejs fix: weavejs uses the plain string 'type' for now.
        if d == "type":
            return cls()
        res = {}
        for k, t in d.items():
            if k != "type" and not k.startswith("_"):
                res[k] = TypeRegistry.type_from_dict(t)
        return cls(res)


@dataclasses.dataclass(frozen=True)
class Function(Type):
    name = "function"

    instance_classes: typing.ClassVar[typing.List[type]] = []

    input_types: dict[str, Type] = dataclasses.field(default_factory=dict)
    output_type: Type = dataclasses.field(default_factory=lambda: UnknownType())

    def _hashable(self):
        return (
            tuple((k, hash(v)) for k, v in self.input_types.items()),
            hash(self.output_type),
        )

    def _is_assignable_to(self, other_type) -> typing.Optional[bool]:
        if other_type.__class__ != Function:
            return other_type.assign_type(self.output_type)
        return None

    def assign_type(self, next_type: Type) -> bool:
        if not self.input_types and not isinstance(next_type, Function):
            # Allow assignment of T to () -> T.
            # Compile handles this in the compile_quote pass.
            return self.output_type.assign_type(next_type)
        return super().assign_type(next_type)

    @classmethod
    def type_of_instance(cls, obj):
        if isinstance(obj.type, Function):
            # Its already a Function type, so just return it.
            # TODO: I'm not sure if this is right, this makes FunctionType
            # a top of sorts...
            return obj.type
        # TODO: get input variable types!
        return cls({}, obj.type)

    @classmethod
    def from_dict(cls, json):
        input_types = {
            pname: TypeRegistry.type_from_dict(ptype)
            for (pname, ptype) in json["inputTypes"].items()
        }
        return cls(input_types, TypeRegistry.type_from_dict(json["outputType"]))

    def _to_dict(self):
        input_types = {k: v.to_dict() for (k, v) in self.input_types.items()}
        return {"inputTypes": input_types, "outputType": self.output_type.to_dict()}

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(obj.to_json(), f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.object.json") as f:
            # TODO: no circular imports!
            from . import graph

            return graph.Node.node_from_json(json.load(f))


@dataclasses.dataclass(frozen=True)
class RefType(Type):
    object_type: Type = UnknownType()

    @classmethod
    def type_of_instance(cls, obj):
        return cls(obj.type)

    # TODO: Address this comment. I'm sure this introduced the same type
    #     blowup as before again. Slows everything down. But in this PR
    #     I've put object_type back into RefType for now.
    # RefType intentionally does not include the type of the object it
    # points to. Doing so naively results in a type explosion. We may
    # want to include it in the future, but we'll need to shrink it, for
    # example by disallowing unions (replacing with them with unknown or a
    # type variable)
    # def __repr__(self):
    #     return "<%s>" % self.__class__.name


@dataclasses.dataclass(frozen=True)
class FilesystemArtifactRefType(RefType):
    pass


@dataclasses.dataclass(frozen=True)
class LocalArtifactRefType(FilesystemArtifactRefType):
    pass


@dataclasses.dataclass(frozen=True)
class WandbArtifactRefType(FilesystemArtifactRefType):
    pass


class WBTable(Type):
    name = "wbtable"


def is_json_compatible(type_):
    if isinstance(type_, List):
        return is_json_compatible(type_.object_type)
    elif isinstance(type_, Dict):
        return is_json_compatible(type_.object_type)
    elif isinstance(type_, TypedDict):
        return all(t for t in type_.property_types.values())
    else:
        return isinstance(type_, BasicType)


def optional(type_: Type) -> Type:
    # TODO: This is not a complete implementation, it doesn't handle
    # Const for example. Although why would we have a Const optional,
    # that doesn't make sense (if its a const, we know the concrete type,
    # its not a union).
    return union(none_type, type_)


def is_optional(type_: Type) -> bool:
    TaggedValueType = type_name_to_type("tagged")

    if isinstance(type_, Const):
        return is_optional(type_.val_type)

    if isinstance(type_, TaggedValueType):
        return is_optional(type_.value)

    return isinstance(type_, UnionType) and any(
        (none_type.assign_type(m)) for m in type_.members
    )


def simple_non_none(type_: Type) -> Type:
    if not isinstance(type_, UnionType):
        return type_
    new_members = [m for m in type_.members if m != NoneType()]
    return union(*new_members)


def non_none(type_: Type) -> Type:
    TaggedValueType = type_name_to_type("tagged")

    if type_ == none_type:
        return Invalid()

    if isinstance(type_, Const):
        val_type = non_none(type_.val_type)
        if val_type == type_.val_type:
            return type_
        return val_type

    if isinstance(type_, TaggedValueType):
        return TaggedValueType(type_.tag, non_none(type_.value))

    if is_optional(type_):
        type_ = typing.cast(UnionType, type_)
        new_members = [
            non_none(m) for m in type_.members if not none_type.assign_type(m)
        ]
        # Can happen if there are Nones and Tagged Nones
        if len(new_members) == 0:
            return Invalid()
        elif len(new_members) == 1:
            return new_members[0]
        else:
            return union(*new_members)
    return type_


def string_enum_type(*vals):
    return UnionType(*[Const(String(), v) for v in vals])


def literal(val: typing.Any) -> Const:
    return Const(TypeRegistry.type_of(val), val)


RUN_STATE_TYPE = string_enum_type("pending", "running", "finished", "failed")


# TODO: get rid of all the underscores. This is another
#    conflict with making ops automatically lazy
@dataclasses.dataclass(frozen=True)
class RunType(ObjectType):
    id: Type = String()
    inputs: Type = TypedDict({})
    history: Type = List(TypedDict({}))
    output: Type = Any()

    def property_types(self):
        return {
            "id": String(),
            "op_name": String(),
            "state": RUN_STATE_TYPE,
            "prints": List(String()),
            "inputs": self.inputs,
            "history": self.history,
            "output": self.output,
        }

    def _is_assignable_to(self, other_type) -> typing.Optional[bool]:
        # Use issubclass, we have RunLocalType defined as a subclass of RunType
        if not issubclass(other_type.__class__, RunType):
            if self.output == NoneType():
                # If output is None, we don't want to be assignable to basically
                # all ops (since ops are nullable)
                return None
            return other_type.assign_type(self.output)
        return None


def merge_many_types(types: list[Type]) -> Type:
    if len(types) == 0:
        return Invalid()
    if len(types) == 1:
        return types[0]
    t = types[0]
    for t2 in types[1:]:
        t = merge_types(t, t2)
    return t


def merge_types(a: Type, b: Type) -> Type:
    """Compute the next list object type.

    list[a].append(b) -> merge_types(a, b)

    This is used at run-time, to figure out the type of existing objects!
    So we can do things like drop UnknownType (UnknownType doesn't make
    sense at run time)

    Note: we don't guarantee that the resulting type can be assigned to
    by either of the inputs! Consider TypedDict[{"a": int}].merge_types(TypedDict[{"b": int}])
    The result is TypedDict[{"a": optional[int], "b": optional[int]}],
    but neither of the inputs can be assigned to it.

    Our decisions about how to merge types here are made to reduce overall
    type size for lists.

    This implementation must match list.concat implementations (which is the only
    way to extend a list in Weave). Ie list.concat(list[a], [b]) -> list[merge_types(a, b)]
    """
    from .language_features.tagging import tagged_value_type

    if a == b:
        return a
    if isinstance(a, Number) and isinstance(b, Number):
        if a.__class__ == Number or b.__class__ == Number:
            return Number()
        if a.__class__ == Float or b.__class__ == Float:
            return Float()
        return Int()
    if isinstance(a, tagged_value_type.TaggedValueType) and isinstance(
        b, tagged_value_type.TaggedValueType
    ):
        merged_tag_type: TypedDict = typing.cast(TypedDict, merge_types(a.tag, b.tag))
        merged_value_type: Type = merge_types(a.value, b.value)
        return tagged_value_type.TaggedValueType(merged_tag_type, merged_value_type)
    if isinstance(a, TypedDict) and isinstance(b, TypedDict):
        all_keys_dict = {}
        for k in a.property_types.keys():
            all_keys_dict[k] = True
        for k in b.property_types.keys():
            all_keys_dict[k] = True

        next_prop_types = {}
        for key in all_keys_dict.keys():
            self_prop_type = a.property_types.get(key, none_type)
            other_prop_type = b.property_types.get(key, none_type)
            next_prop_types[key] = merge_types(self_prop_type, other_prop_type)
        return TypedDict(next_prop_types)
    if isinstance(a, ObjectType) and isinstance(b, ObjectType):
        if a.name == b.name:
            next_type_attrs = {}
            for key in a.type_attrs():
                next_type_attrs[key] = merge_types(getattr(a, key), getattr(b, key))
            return type(a)(**next_type_attrs)

    if isinstance(a, List) and isinstance(b, List):
        return List(merge_types(a.object_type, b.object_type))

    if isinstance(a, UnknownType):
        return b
    if isinstance(b, UnknownType):
        return a

    if isinstance(a, UnionType) or isinstance(b, UnionType):
        new_union = UnionType(a, b)
        final_members = [new_union.members[0]]
        for mem in new_union.members[1:]:
            for i, final_mem in enumerate(final_members):
                merged_m = merge_types(final_mem, mem)
                if not isinstance(merged_m, UnionType):
                    final_members[i] = merged_m
                    break
            else:
                final_members.append(mem)

        # for i in range(len(new_union.members)):
        #     for j in range(i + 1, len(new_union.members)):
        #         merged_m = merge_types(new_union.members[i], new_union.members[j])
        #         if not isinstance(merged_m, UnionType):
        #             final_members.append(merged_m)
        #         else:
        #             final_members.append(new_union.members[i])
        return union(*final_members)
    return union(a, b)


def unknown_coalesce(in_type: Type) -> Type:
    """
    Recursively removes unknowns from a type. Whenever a union is encountered,
    we do 2 operations:
    1) reduce the union members to only non-unknowns
    2) for each container-type member, attempt to remove any unknowns that are
        known by a peer member.

    For example:

    Union<
        Unknown,
        List<Union<Unknown, Int>>,
        TypedDict<{'a': Unknown}>,
        TypedDict<{'a': Int}>,
    >

    would reduce to

    Union<
        List<Int>,
        TypedDict<{'a': Int}>,
    >

    This is particularly useful when a nested field in an AWL is empty - the
    type will be Unknown. However, when merging for concatenation, we want to
    remove such unknowns since we know the type from a another list.
    """

    # In this function, there are 3 cases:
    # 1) in_type is a union type - in this case we delegate to _unknown_coalesce_on_union
    # 2) Each of the container types calls this function recursively
    # 3) the base case just returns the type
    TaggedValueType = type_name_to_type("tagged")
    if isinstance(in_type, UnionType):
        return _unknown_coalesce_on_union(in_type)
    elif isinstance(in_type, UnknownType):
        return UnknownType()
    elif isinstance(in_type, List):
        return List(unknown_coalesce(in_type.object_type))
    elif isinstance(in_type, Dict):
        return Dict(in_type.key_type, unknown_coalesce(in_type.object_type))
    elif isinstance(in_type, TypedDict):
        return TypedDict(
            {
                key: unknown_coalesce(value_type)
                for key, value_type in in_type.property_types.items()
            }
        )
    elif isinstance(in_type, TaggedValueType):
        return TaggedValueType(
            unknown_coalesce(in_type.tag),  # type: ignore
            unknown_coalesce(in_type.value),
        )
    elif isinstance(in_type, Const):
        return Const(unknown_coalesce(in_type.val_type), in_type.value)
    elif isinstance(in_type, ObjectType):
        # TODO: This is hard due to image type being non-standard.
        pass
    return in_type


def _filter_unknowns(types: list[Type]) -> list[Type]:
    return [t for t in types if not isinstance(t, UnknownType)]


def _unknown_coalesce_on_union(u_type: UnionType) -> Type:
    # This helper function will remove all the unknowns from a a union's members,
    # then apply the `unknown_coalesce` function to each known member. This ensures
    # that each member is processed before comparing to it's peers.
    known_members = [
        unknown_coalesce(member) for member in _filter_unknowns(u_type.members)
    ]

    if len(known_members) == 0:
        return UnknownType()

    final_types: list[Type] = []
    for ndx in range(len(known_members)):
        final_types.append(
            _merge_unknowns_of_type_with_types(
                known_members[ndx], known_members[:ndx] + known_members[ndx + 1 :]
            )
        )

    return union(*final_types)


def _merge_unknowns_of_type_with_types(of_type: Type, with_types: list[Type]):
    # At this stage, we have a starting type and a list of peer candidate types.
    # These types have already been coalesced individually. This step is similar to
    # the outer `unknown_coalesce` function, except that rather than operating on the
    # single incoming type, we must select the correct type from the list of peers.
    TaggedValueType = type_name_to_type("tagged")

    # First, we filter out any unknowns from the list of peers.
    with_types = _filter_unknowns(with_types)
    if len(with_types) == 0:
        return of_type

    # If the current type itself is a union, then we need to recurse into it
    if isinstance(of_type, UnionType):
        return union(
            *[
                _merge_unknowns_of_type_with_types(member, with_types)
                for member in of_type.members
            ]
        )

    # if the current type is unknown, then we just return the next peer type.
    elif isinstance(of_type, UnknownType):
        return with_types[0]

    # Next, we filter down to peer types that have the same type class, and recurse
    # into each container type.
    with_types = [
        member for member in with_types if member.__class__ == of_type.__class__
    ]
    if isinstance(of_type, List):
        return List(
            _merge_unknowns_of_type_with_types(
                of_type.object_type, [t.object_type for t in with_types]  # type: ignore
            )
        )
    elif isinstance(of_type, Dict):
        return Dict(
            of_type.key_type,
            _merge_unknowns_of_type_with_types(
                of_type.value_type, [t.value_type for t in with_types]  # type: ignore
            ),
        )
    elif isinstance(of_type, TypedDict):
        return TypedDict(
            {
                key: _merge_unknowns_of_type_with_types(
                    value_type, [t.property_types.get(key, NoneType()) for t in with_types]  # type: ignore
                )
                for key, value_type in of_type.property_types.items()
            }
        )
    elif isinstance(of_type, TaggedValueType):
        return TaggedValueType(
            _merge_unknowns_of_type_with_types(of_type.tag, [t.tag for t in with_types]),  # type: ignore
            _merge_unknowns_of_type_with_types(
                of_type.value, [t.value for t in with_types]  # type: ignore
            ),
        )
    elif isinstance(of_type, Const):
        return Const(
            _merge_unknowns_of_type_with_types(
                of_type.val_type, [t.val_type for t in with_types]  # type: ignore
            ),
            of_type.value,
        )
    elif isinstance(of_type, ObjectType):
        # TODO: This is hard due to image type being non-standard.
        pass

    # Finally, return the type.
    return of_type


def union(*members: Type) -> Type:
    if not members:
        return UnknownType()
    final_members = []
    for member in members:
        if isinstance(member, UnionType):
            for sub_member in member.members:
                if sub_member not in final_members:
                    final_members.append(sub_member)
        elif member not in final_members:
            final_members.append(member)
    # Unknown "takes over" in a union.
    if any(isinstance(m, UnknownType) for m in final_members):
        return UnknownType()
    if len(final_members) == 1:
        return final_members[0]
    return UnionType(*final_members)


def is_list_like(t: Type) -> bool:
    return List().assign_type(non_none(t))


def is_custom_type(t: Type) -> bool:
    return not (
        isinstance(t, BasicType)
        or isinstance(t, ObjectType)
        or isinstance(t, TypedDict)
        or isinstance(t, List)
        or isinstance(t, UnionType)
        or isinstance(t, Timestamp)
        or t.name == "tagged"
    )


def type_is_variable(t: Type) -> bool:
    # not concrete
    if isinstance(t, Any):
        # Any is currently the equivalent Weave Type for any Python
        # TypeVar
        return True
    elif isinstance(t, UnionType):
        return True
    elif isinstance(t, TypedDict):
        # Should we just make type_vars for TypedDict be its property types?
        # That would simplify a lot.
        return any(type_is_variable(sub_t) for sub_t in t.property_types.values())
    elif isinstance(t, Const):
        return type_is_variable(t.val_type)
    else:
        return any(type_is_variable(sub_t) for sub_t in t.type_vars.values())


NumberBinType = TypedDict({"start": Float(), "stop": Float()})
TimestampBinType = TypedDict({"start": Timestamp(), "stop": Timestamp()})


def map_leaf_types(t: Type, fn: typing.Callable[[Type], typing.Optional[Type]]) -> Type:
    """
    This function will recursively apply `fn` to all leaf types in `t`. A leaf type is
    a basic type, object type, or const. If `fn` returns None, then the original type
    will be used.
    """

    def null_safe_fn(t: Type) -> Type:
        fn_res = fn(t)
        return t if fn_res is None else fn_res

    return _map_leaf_types_null_safe(t, null_safe_fn)


def _map_leaf_types_null_safe(t: Type, fn: typing.Callable[[Type], Type]) -> Type:
    TaggedValueType = type_name_to_type("tagged")
    if isinstance(t, List):
        return List(_map_leaf_types_null_safe(t.object_type, fn))
    elif isinstance(t, TypedDict):
        return TypedDict(
            {k: _map_leaf_types_null_safe(v, fn) for k, v in t.property_types.items()}
        )
    elif isinstance(t, Dict):
        return Dict(t.key_type, _map_leaf_types_null_safe(t.object_type, fn))
    elif isinstance(t, ObjectType):
        pass
    elif isinstance(t, UnionType):
        return union(*[_map_leaf_types_null_safe(t, fn) for t in t.members])
    elif isinstance(t, Const):
        pass
    elif isinstance(t, TaggedValueType):
        return TaggedValueType(
            _map_leaf_types_null_safe(t.tag, fn),  # type: ignore
            _map_leaf_types_null_safe(t.value, fn),
        )
    return fn(t)


# parse_const_type and const_type_to_json are a new concept used only
# by the weave0 ImageArtifactRefFileType. That stores for example
# boxScores as {box1: [2, 3, 5], box2: [0]} where the lists are
# class ids. This is a way to convert those literal representations
# to types, making assignability work. The above becomes
# TypedDict<box1: List<Union<Const<2>, Const<3>, Const<5>>>, box2: List<Const<0>>>.
# We can update the names and standardize this more if we find it useful
# outside of weave0 support.


def parse_constliteral_type(val: typing.Any) -> Type:
    if isinstance(val, Type):
        return val
    elif isinstance(val, dict):
        return TypedDict({k: parse_constliteral_type(v) for k, v in val.items()})
    elif isinstance(val, list):
        return List(union(*[parse_constliteral_type(v) for v in val]))
    else:
        return Const(TypeRegistry.type_of(val), val)


def constliteral_type_to_json(t: Type) -> typing.Any:
    t = simple_non_none(t)
    if isinstance(t, Const):
        return t.val
    elif isinstance(t, TypedDict):
        return {k: constliteral_type_to_json(v) for k, v in t.property_types.items()}
    elif isinstance(t, Dict):
        return {}
    elif isinstance(t, List):
        object_type = t.object_type
        members = [object_type]
        if isinstance(object_type, UnionType):
            members = object_type.members
        members = [m for m in members if isinstance(m, Const)]
        return [constliteral_type_to_json(mem) for mem in members]
    raise ValueError(f"Cannot convert {t} to a JSON value")


def split_none(t: Type) -> tuple[bool, Type]:
    """Returns (is_optional, non_none_type)"""
    if t == NoneType():
        return True, UnknownType()
    if not isinstance(t, UnionType):
        return False, t
    non_none_members = [m for m in t.members if not isinstance(m, NoneType)]
    if len(non_none_members) < len(t.members):
        return True, union(*non_none_members)
    return False, t



import pathlib
import typing

from . import errors


def safe_join(*args: typing.Union[str, pathlib.Path]) -> str:
    # Ensure result is relative to first dir.
    first_part = pathlib.Path(args[0])
    result = first_part.joinpath(*args[1:])
    if not result.resolve().is_relative_to(first_part.resolve()):
        raise errors.WeaveAccessDeniedError()
    return str(result)



import datetime
import itertools
import random
import threading
import time
import weave
from weave.timestamp import tz_aware_dt


@weave.type()
class Prediction:
    timestamp: datetime.datetime
    input: float
    output: float


@weave.type()
class EvalModel:
    id: str

    @weave.op(pure=False)
    def all_predictions(
        self, start_datetime: datetime.datetime, end_datetime: datetime.datetime
    ) -> list[Prediction]:
        pred_procs_node = weave.ops.objects(
            PredictionProcess.WeaveType(), "latest", 0  # type: ignore
        ).filter(lambda proc: proc.get().model.id == self.id)
        # TODO: this'll be slow! Should all be done in weave
        # TODO: use ref metadata to determine if a given obj falls
        #    in the window. (or lazily load refs and put metadata on PredProc)
        pred_procs = weave.use(pred_procs_node)
        res = list(itertools.chain(*(p.get().predictions for p in pred_procs)))
        return [
            p
            for p in res
            if p.timestamp > tz_aware_dt(start_datetime)
            and p.timestamp < tz_aware_dt(end_datetime)
        ]


@weave.type()
class PredictionProcess:
    id: int
    model: EvalModel
    predictions: list[Prediction]


class Predictor(threading.Thread):
    def __init__(self, model_id, run_for_s):
        threading.Thread.__init__(self)
        self.model = EvalModel(model_id)
        self.prediction_process = PredictionProcess(
            random.randrange(0, 10000000), self.model, []
        )
        self.run_for_s = run_for_s

    def run(self):
        start_time = time.time()
        while time.time() - start_time < self.run_for_s:
            for i in range(50):
                self.prediction_process.predictions.append(
                    Prediction(random.random(), random.random())
                )
                time.sleep(0.1)
            weave.save(
                self.prediction_process,
                name=f"model_preds-{self.model.id}-{self.prediction_process.id}",
            )


# Data model
# run.log()



import pytest

from .. import api as weave
from .. import weave_types as types
from .. import box
from . import list_
from . import dict
from . import number
from . import runs
from . import errors
from .. import weave_internal
from ..tests import weavejs_ops
from ..tests import geom
from ..language_features.tagging import make_tag_getter_op, tag_store, tagged_value_type


def test_unnest():
    data = [{"a": [1, 2], "b": "x", "c": [5, 6]}, {"a": [3, 4], "b": "j", "c": [9, 10]}]
    unnested = weave.use(list_.unnest(data))
    # convert to list so pytest prints nice diffs if there is a mismatch
    assert list(unnested) == [
        {"a": 1, "b": "x", "c": 5},
        {"a": 2, "b": "x", "c": 6},
        {"a": 3, "b": "j", "c": 9},
        {"a": 4, "b": "j", "c": 10},
    ]


def test_op_list():
    node = list_.make_list(a=1, b=2, c=3)
    assert types.List(types.Int()).assign_type(node.type)


def test_sequence1():
    l = [
        {"a": 0, "b": 0, "y": "x"},
        {"a": 0, "b": 0, "y": "y"},
        {"a": 0, "b": 1, "y": "x"},
        {"a": 0, "b": 1, "y": "y"},
        {"a": 0, "b": 1, "y": "y"},
        {"a": 1, "b": 1, "y": "y"},
        {"a": 2, "b": 1, "y": "y"},
    ]
    saved = weave.save(l)

    # PanelFacet GroupBy
    groupby1_fn = weave.define_fn(
        {"row": types.TypeRegistry.type_of(l).object_type},
        lambda row: dict.dict_(
            **{
                "a": row["a"],
                "b": row["b"],
            }
        ),
    )
    res = weavejs_ops.groupby(saved, groupby1_fn)

    # Input to PanelPlot
    map1_fn = weave.define_fn({"row": res.type.object_type}, lambda row: row)
    res = weavejs_ops.map(res, map1_fn)

    inner_groupby_fn = weave.define_fn(
        {"row": res.type.object_type.object_type},
        lambda row: dict.dict_(**{"y": row["y"]}),
    )
    map2_fn = weave.define_fn(
        {"row": res.type.object_type},
        lambda row: row.groupby(inner_groupby_fn).map(
            weave.define_fn(
                {
                    "row": tagged_value_type.TaggedValueType(
                        types.TypedDict(
                            {"groupKey": types.TypedDict({"y": types.String()})}
                        ),
                        res.type.object_type,
                    )
                },
                lambda row: row.groupkey()["y"],
            )
        ),
    )
    res = weavejs_ops.map(res, map2_fn)
    res = list_.flatten(res)
    res = list_.unique(res)
    assert list(weave.use(res)) == ["x", "y"]


def test_nested_functions():
    rows = weave.save([{"a": [1, 2]}])
    map_fn = weave.define_fn(
        {"row": types.TypedDict({"a": types.List(types.Int())})},
        lambda row: number.numbers_avg(
            row["a"].map(weave.define_fn({"row": types.Int()}, lambda row: row + 1))
        ),
    )
    mapped = rows.map(map_fn)
    assert weave.use(mapped) == [2.5]


def test_concat():
    list_node_1 = list_.make_list(a=1, b=2, c=3)
    list_node_2 = list_.make_list(a=10, b=20, c=30)
    list_node_3 = list_.make_list(a=list_node_1, b=list_node_2)
    concat_list = list_node_3.concat()
    assert weave.use(concat_list) == [1, 2, 3, 10, 20, 30]


def test_nullable_concat():
    list_node_1 = list_.make_list(a=1, b=2, c=3)
    list_node_2 = list_.make_list(a=10, b=20, c=30)
    list_node_3 = list_.make_list(a=list_node_1, b=list_node_2, c=weave.save(None))
    concat_list = list_node_3.concat()
    assert weave.use(concat_list) == [1, 2, 3, 10, 20, 30]


def test_mapeach():
    row_node = list_.make_list(a=1, b=2, c=3)
    two_d_list_node = list_.make_list(a=row_node, b=row_node, c=row_node)
    result = list_.map_each(two_d_list_node, lambda row: row + 1)
    assert result.type == types.List(types.List(types.Number()))
    assert weave.use(result) == [[2, 3, 4], [2, 3, 4], [2, 3, 4]]


def test_mapeach_single():
    row_node = list_.make_list(a=1, b=2, c=3)
    item = row_node[0]
    result = list_.map_each(item, lambda row: row + 1)
    assert result.type == types.Number()
    assert weave.use(result) == 2


def test_mapeach_tagged():
    raw_data = [[i + 3 * j for i in [2, 3, 4]] for j in [0, 1, 2]]
    for i, row in enumerate(raw_data):
        for j, value in enumerate(row):
            raw_data[i][j] = tag_store.add_tags(box.box(value), {"row": i, "col": j})
        raw_data[i] = tag_store.add_tags(box.box(raw_data[i]), {"row": i})
    raw_data = tag_store.add_tags(box.box(raw_data), {"tag": "top"})
    two_d_list_node = weave.save(raw_data)

    result = list_.map_each(two_d_list_node, lambda row: row + 1)

    expected_type = tagged_value_type.TaggedValueType(
        types.TypedDict({"tag": types.String()}),
        types.List(
            tagged_value_type.TaggedValueType(
                types.TypedDict({"row": types.Int()}),
                types.List(
                    tagged_value_type.TaggedValueType(
                        types.TypedDict({"row": types.Int(), "col": types.Int()}),
                        types.Number(),
                    )
                ),
            )
        ),
    )

    assert expected_type.assign_type(result.type)
    assert weave.use(result) == [[i + 3 * j for i in [3, 4, 5]] for j in [0, 1, 2]]

    tag_getter_op_top = make_tag_getter_op.make_tag_getter_op("tag", types.String())
    tag_getter_op_row = make_tag_getter_op.make_tag_getter_op("row", types.Int())
    tag_getter_op_col = make_tag_getter_op.make_tag_getter_op("col", types.Int())

    assert weave.use(tag_getter_op_top(result)) == "top"
    for i in range(3):
        assert weave.use(tag_getter_op_row(result[i])) == i
        for j in range(3):
            assert weave.use(tag_getter_op_row(result[i][j])) == i
            assert weave.use(tag_getter_op_col(result[i][j])) == j


def test_dropna_tagged():
    raw_data = [1, None, 3]
    for i, item in enumerate(raw_data):
        raw_data[i] = tag_store.add_tags(box.box(item), {"row": i})
    list_node = weave.save(raw_data)
    dropnaed = list_node.dropna()

    assert weave.use(dropnaed) == [1, 3]

    tag_getter_op_row = make_tag_getter_op.make_tag_getter_op("row", types.Int())

    for i, val in enumerate([0, 2]):
        assert weave.use(tag_getter_op_row(dropnaed[i])) == val


def test_preserve_tags_on_mapped_elements():
    data = [1]
    for i, item in enumerate(data):
        data[i] = tag_store.add_tags(box.box(item), {"row": i})
    list_node = weave.save(data)
    mapped = list_node.map(lambda n: n + 2)
    indexed = mapped[0]
    tag_getter_op = make_tag_getter_op.make_tag_getter_op("row", types.Int())
    assert weave.use(tag_getter_op(indexed)) == 0


def test_group_over_tagged():
    l = [
        tag_store.add_tags(box.box({"a": 0, "b": 0, "y": "x"}), {"row": 0}),
        tag_store.add_tags(box.box({"a": 0, "b": 1, "y": "x"}), {"row": 1}),
        tag_store.add_tags(box.box({"a": 0, "b": 0, "y": "x"}), {"row": 2}),
        tag_store.add_tags(box.box({"a": 1, "b": 1, "y": "x"}), {"row": 3}),
        tag_store.add_tags(box.box({"a": 1, "b": 0, "y": "x"}), {"row": 4}),
        tag_store.add_tags(box.box({"a": 1, "b": 1, "y": "x"}), {"row": 5}),
    ]
    saved = weave.save(l)

    groupby1_fn = weave.define_fn(
        {"row": types.TypeRegistry.type_of(l).object_type},
        lambda row: dict.dict_(
            **{
                "a": row["a"],
            }
        ),
    )
    res = weavejs_ops.groupby(saved, groupby1_fn)
    assert weave.use(res) == [
        [
            {"a": 0, "b": 0, "y": "x"},
            {"a": 0, "b": 1, "y": "x"},
            {"a": 0, "b": 0, "y": "x"},
        ],
        [
            {"a": 1, "b": 1, "y": "x"},
            {"a": 1, "b": 0, "y": "x"},
            {"a": 1, "b": 1, "y": "x"},
        ],
    ]


def test_lookup():
    l = weave.save([{"id": 5, "b": 5}, {"id": 6, "b": 6}])
    assert weave.use(l.lookup(0)) == None
    assert weave.use(l.lookup(5)) == {"id": 5, "b": 5}

    # can't call on TypedDict with no id
    l = weave.save([{"b": 5}, {"b": 6}])
    with pytest.raises(errors.WeaveDispatchError):
        l.lookup(0)

    l = weave.save(
        [runs.Run("a", "cool-op", output=25), runs.Run("b", "cool-op", output=26)]
    )
    assert weave.use(l.lookup(0)) == None
    assert weave.use(l.lookup("a")) == runs.Run("a", "cool-op", output=25)

    # Can't call on object with no ID
    l = weave.save([geom.Point2d(0, 5)])
    with pytest.raises(errors.WeaveDispatchError):
        l.lookup(0)


def test_cross_product():
    obj = weave.save({"a": ["x", "y", "z"], "b": [1, 2, 3]})
    cp = list_.cross_product(obj)
    assert cp.type == weave.types.List(
        object_type=weave.types.TypedDict(
            property_types={"a": weave.types.String(), "b": weave.types.Int()}
        )
    )
    assert weave.use(cp) == [
        {"a": "x", "b": 1},
        {"a": "x", "b": 2},
        {"a": "x", "b": 3},
        {"a": "y", "b": 1},
        {"a": "y", "b": 2},
        {"a": "y", "b": 3},
        {"a": "z", "b": 1},
        {"a": "z", "b": 2},
        {"a": "z", "b": 3},
    ]



import os
import pathlib
import typing

from . import environment
from . import path_util
from . import cache


def get_allowed_dir() -> pathlib.Path:
    if not environment.is_public():
        return pathlib.Path("/")
    cache_namespace = cache.get_user_cache_key()
    if cache_namespace is None:
        raise ValueError("cache_namespace is None but is_public() is True")
    return pathlib.Path(environment.weave_filesystem_dir()) / cache_namespace


def path_ext(path: str) -> str:
    return os.path.splitext(path)[1].strip(".")


def check_path(path: str) -> None:
    allowed_dir = get_allowed_dir()
    path_util.safe_join(allowed_dir, path)


def safe_open(path: str, mode: str = "r") -> typing.IO:
    check_path(path)
    # ensure that the directory exists
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return open(path, mode)



# This demonstrates how to create asynchronous Weave ops.

import time

import weave


@weave.op(
    render_info={"type": "function"},
    name="demo-slowmult",
    input_type={
        "a": weave.types.Int(),
        "b": weave.types.Int(),
        "sleep": weave.types.Float(),
    },
    output_type=weave.types.Function(
        output_type=weave.types.RunType(
            output=weave.types.Int(),
        )
    ),
)
def slowmult(a, b, sleep, _run=None):
    res = b
    for i in range(a - 1):
        res += b
        _run.print_("Current result %s" % res)
        time.sleep(sleep)
    _run.set_output(res)


class AsyncDemoModelType(weave.types.ObjectType):
    name = "asyncdemo_model"

    def __init__(self):
        pass

    def property_types(self):
        return {"id": weave.types.String()}


@weave.weave_class(weave_type=AsyncDemoModelType)
class AsyncDemoModel:
    def __init__(self, id):
        self.id = id

    @weave.op(
        name="asyncdemo_model-infer",
        input_type={"self": AsyncDemoModelType(), "x": weave.types.String()},
        output_type=weave.types.String(),
    )
    def infer(self, x):
        return x + "-abcd"


AsyncDemoModelType.instance_classes = AsyncDemoModel
AsyncDemoModelType.instance_class = AsyncDemoModel


class AsyncDemoTrainResultType(weave.types.ObjectType):
    name = "asyncdemo_trainresult"

    def __init__(self):
        pass

    def property_types(self):
        return {"id": weave.types.String()}


@weave.weave_class(weave_type=AsyncDemoTrainResultType)
class AsyncDemoTrainResult:
    def __init__(self, id):
        self.id = id

    @weave.op(
        name="asyncdemo_trainresult-model",
        input_type={"self": AsyncDemoTrainResultType()},
        output_type=AsyncDemoModelType(),
    )
    def model(self):
        return AsyncDemoModel(self.id)


AsyncDemoTrainResultType.instance_classes = AsyncDemoTrainResult
AsyncDemoTrainResultType.instance_class = AsyncDemoTrainResult


@weave.op(
    render_info={"type": "function"},
    name="demo-train",
    input_type={
        "dataset": weave.types.List(
            weave.types.TypedDict(
                {"prompt": weave.types.String(), "completion": weave.types.String()}
            )
        ),
    },
    output_type=weave.types.Function(
        output_type=weave.types.RunType(
            output=AsyncDemoTrainResultType(),
        )
    ),
)
def train(dataset, _run=None):
    _run.print_("starting")
    _run.set_output(AsyncDemoTrainResult(dataset[0]["prompt"]))
    _run.print_("done")



from .. import api as weave
from .. import box
from . import any

import pytest
import typing


cases: list[typing.Tuple[typing.Any, bool]] = [
    (1, False),
    (None, True),
    ([], False),
    (box.box(None), True),
    ([1, 2, 3], False),
]


@pytest.mark.parametrize(["input", "expected_output"], cases)
def test_isnone(input, expected_output):
    node = any.is_none(input)
    assert weave.use(node) == expected_output



import random
import pyarrow as pa
import pyarrow.compute as pc
import numpy as np
import typing
from builtins import map as builtin_map

from ..api import op, type_of
from ..decorator_arrow_op import arrow_op
from .. import weave_types as types
from ..language_features.tagging import (
    tagged_value_type,
    tagged_value_type_helpers,
)
from collections import defaultdict
from .. import op_args
from ..ops_primitives import list_ as primitive_list
from .. import op_def

from .arrow import ArrowWeaveListType, arrow_as_array, offsets_starting_at_zero
from .list_ import ArrowWeaveList, PathType, is_list_arrowweavelist
from . import arrow_tags
from .vectorize import _apply_fn_node_with_tag_pushdown
from . import convert
from .convert import to_compare_safe
from .constructors import (
    vectorized_container_constructor_preprocessor,
    vectorized_input_types,
)


FLATTEN_DELIMITER = "➡️"

NestedTableColumns = dict[str, typing.Union[dict, pa.ChunkedArray]]


def unflatten_structs_in_flattened_table(table: pa.Table) -> pa.Table:
    """take a table with column names like {a.b.c: [1,2,3], a.b.d: [4,5,6], a.e: [7,8,9]}
    and return a struct array with the following structure:
    [ {a: {b: {c: 1, d: 4}, e: 7}}, {a: {b: {c: 2, d: 5}, e: 8}}, {a: {b: {c: 3, d: 6}, e: 9}} ]
    """

    def recursively_build_nested_struct_array(
        columns: dict[str, pa.Array]
    ) -> pa.StructArray:
        result: NestedTableColumns = defaultdict(lambda: {})
        for colname in columns:
            spl_colname = colname.split(FLATTEN_DELIMITER)
            prefix = spl_colname[0]
            suffix = FLATTEN_DELIMITER.join(spl_colname[1:])
            if suffix:
                result[prefix][suffix] = columns[colname]
            else:
                result[prefix] = columns[colname]

        for colname in result:
            if isinstance(result[colname], dict):
                result[colname] = recursively_build_nested_struct_array(result[colname])

        names: list[str] = []
        arrays: list[pa.ChunkedArray] = []

        for name, array in result.items():
            names.append(name)
            arrays.append(array)

        return pa.StructArray.from_arrays(arrays, names)

    recurse_input = {
        colname: table[colname].combine_chunks() for colname in table.column_names
    }

    sa = recursively_build_nested_struct_array(recurse_input)

    names: list[str] = []
    chunked_arrays: list[pa.ChunkedArray] = []

    for field in sa.type:
        names.append(field.name)
        chunked_arrays.append(pa.chunked_array(sa.field(field.name)))

    return pa.Table.from_arrays(chunked_arrays, names=names)


def _recursively_flatten_structs_in_array(
    arr: pa.Array, prefix: str, _stack_depth=0
) -> dict[str, pa.Array]:
    if pa.types.is_struct(arr.type):
        result: dict[str, pa.Array] = {}
        for field in arr.type:
            new_prefix = (
                prefix + (FLATTEN_DELIMITER if _stack_depth > 0 else "") + field.name
            )
            result.update(
                _recursively_flatten_structs_in_array(
                    arr.field(field.name),
                    new_prefix,
                    _stack_depth=_stack_depth + 1,
                )
            )
        return result
    return {prefix: arr}


def unzip_struct_array(arr: pa.ChunkedArray) -> pa.Table:
    flattened = _recursively_flatten_structs_in_array(arr.combine_chunks(), "")
    return pa.table(flattened)


@op(
    name="ArrowWeaveList-map",
    input_type={
        "self": ArrowWeaveListType(),
        "map_fn": lambda input_types: types.Function(
            {"row": input_types["self"].object_type, "index": types.Int()},
            types.Any(),
        ),
    },
    output_type=lambda input_types: ArrowWeaveListType(
        input_types["map_fn"].output_type
    ),
)
def map(self, map_fn):
    res = _apply_fn_node_with_tag_pushdown(self, map_fn)
    return res


def _map_each_function_type(input_types: dict[str, types.Type]) -> types.Type:
    base_op_fn_type = primitive_list._map_each_function_type(
        {"arr": input_types["self"]}
    )
    base_op_fn_type.input_types["self"] = ArrowWeaveListType(
        typing.cast(types.List, input_types["self"]).object_type
    )
    return base_op_fn_type


def _map_each_output_type(input_types: dict[str, types.Type]):
    base_output_type = primitive_list._map_each_output_type(
        {"arr": input_types["self"], "mapFn": input_types["map_fn"]}
    )
    return ArrowWeaveListType(base_output_type.object_type)


def _map_each(self: ArrowWeaveList, fn):
    if types.List().assign_type(self.object_type):
        as_array = arrow_as_array(self._arrow_data)
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            as_array = as_array.field("_value")

        offsets = offsets_starting_at_zero(as_array)
        flattened = as_array.flatten()
        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            new_object_type = typing.cast(
                types.List, self.object_type.value
            ).object_type
        else:
            new_object_type = typing.cast(types.List, self.object_type).object_type

        new_awl: ArrowWeaveList = ArrowWeaveList(
            flattened,
            new_object_type,
            self._artifact,
        )

        mapped = _map_each(new_awl, fn)
        reshaped_arrow_data: ArrowWeaveList = ArrowWeaveList(
            pa.ListArray.from_arrays(offsets, mapped._arrow_data),
            types.List(mapped.object_type),
            self._artifact,
        )

        if isinstance(self.object_type, tagged_value_type.TaggedValueType):
            return arrow_tags.awl_add_arrow_tags(
                reshaped_arrow_data,
                self._arrow_data.field("_tag"),
                self.object_type.tag,
            )
        return reshaped_arrow_data
    res = _apply_fn_node_with_tag_pushdown(self, fn)
    return res


@op(
    name="ArrowWeaveList-mapEach",
    input_type={
        "self": ArrowWeaveListType(),
        "map_fn": _map_each_function_type,
    },
    output_type=_map_each_output_type,
)
def map_each(self, map_fn):
    return _map_each(self, map_fn)


def _impute_nones_for_sort(awl: ArrowWeaveList) -> ArrowWeaveList:
    """Impute missing values in a table so that it can be sorted. This is a workaround for
    the fact that arrow does not handle missing values in a table when sorting, it just
    moves them all to the end of the list."""

    def impute_nones(
        arrow_list: ArrowWeaveList, path: PathType
    ) -> typing.Optional[ArrowWeaveList]:
        # this does not work for unions in arrow < 12
        if not types.is_optional(arrow_list.object_type):
            return None

        mask = pa.compute.is_null(arrow_list._arrow_data)
        any_null = pa.compute.any(mask).as_py()
        num_null = pa.compute.sum(mask).as_py()

        if any_null:
            all_null = num_null == len(arrow_list._arrow_data)
            if all_null:
                return ArrowWeaveList(
                    pa.repeat(0, num_null),
                    types.Int(),
                    awl._artifact,
                )
            if pa.types.is_struct(arrow_list._arrow_data.type) or pa.types.is_list(
                arrow_list._arrow_data.type
            ):
                raise NotImplementedError(
                    "Imputing nulls for a struct or a list is not implemented"
                )

            if pa.types.is_string(arrow_list._arrow_data.type):
                replacement = pa.repeat("", num_null)
                new_data = pa.compute.replace_with_mask(
                    arrow_list._arrow_data, mask, replacement
                )
            elif pa.types.is_floating(arrow_list._arrow_data.type):
                min = pa.compute.min(arrow_list._arrow_data).as_py()
                replacement = pa.repeat(
                    min - 1.0, num_null
                )  # greater than all other elements
            elif pa.types.is_integer(arrow_list._arrow_data.type):
                min = pa.compute.min(arrow_list._arrow_data).as_py()
                replacement = pa.repeat(min - 1, num_null)
            else:
                raise NotImplementedError(
                    f"Imputing nulls for type {arrow_list._arrow_data.type} is not implemented"
                )

            new_data = pa.compute.replace_with_mask(
                arrow_list._arrow_data, mask, replacement
            )
            return ArrowWeaveList(
                new_data,
                types.non_none(arrow_list.object_type),
                awl._artifact,
            )

        return None

    return awl.map_column(impute_nones)


def _sort_values_list_array_to_table(
    sort_values: ArrowWeaveList, col_dirs: list[str]
) -> pa.Table:
    flattened = sort_values._arrow_data_asarray_no_tags().flatten()

    columns = (
        []
    )  # this is intended to be a pylist and will be small since it is number of sort fields
    col_len = len(sort_values._arrow_data)
    dir_len = len(col_dirs)
    col_names = [str(i) for i in range(dir_len)]
    # arrow data: (col_len x dir_len)
    # [
    #    [d00, d01]
    #    [d10, d11]
    #    [d20, d21]
    #    [d30, d31]
    # ]
    #
    # Flatten
    # [d00, d01, d10, d11, d20, d21, d30, d31]
    #
    # col_x = [d0x, d1x, d2x, d3x]
    # .         i * dir_len + x
    #
    #
    for i in range(dir_len):
        take_array = [j * dir_len + i for j in range(col_len)]
        if len(take_array) > 0:
            columns.append(pc.take(flattened, pa.array(take_array)))
        else:
            columns.append(flattened)

    return pa.Table.from_arrays(columns, names=col_names)


def _arrow_sort_values_to_indices(
    sort_values: pa.Table, col_dirs: list[str]
) -> pa.Array:
    col_names = sort_values.column_names

    order = [
        (col_name, "ascending" if dir_name == "asc" else "descending")
        for col_name, dir_name in zip(col_names, col_dirs)
    ]

    return pc.sort_indices(sort_values, order, null_placement="at_end")


@op(
    name="ArrowWeaveList-sort",
    input_type={
        "self": ArrowWeaveListType(),
        "comp_fn": lambda input_types: types.Function(
            {"row": input_types["self"].object_type, "index": types.Int()},
            types.Any(),
        ),
        "col_dirs": types.List(types.String()),
    },
    output_type=lambda input_types: input_types["self"],
)
def sort(self, comp_fn, col_dirs):
    sort_values = _apply_fn_node_with_tag_pushdown(self, comp_fn)
    sort_values = _sort_values_list_array_to_table(sort_values, col_dirs)

    # modify to_compare_safe to preserve numbers
    for i, column_name in enumerate(sort_values.column_names):
        column = sort_values[column_name]
        column = column.combine_chunks()

        # handle unions.

        if pa.types.is_union(column.type):
            if pa.compute.all(
                pa.compute.equal(column.type_codes, column.type_codes[0])
            ).as_py():
                column = column.field(column.type_codes[0].as_py())
            else:
                # have to break out here. we need to convert to a string but
                # pyarrow doesn't support string casting for multi-type unions.
                column = [str(x) if x is not None else x for x in column.to_pylist()]
                column = pa.array(column)

        # quickly refine column type
        column_type = types.TypeRegistry.type_of(column).object_type
        has_nulls = pa.compute.any(pc.is_null(column)).as_py()
        if has_nulls:
            column_type = types.optional(column_type)

        column = ArrowWeaveList(column, column_type, self._artifact)
        column = _impute_nones_for_sort(column)
        column = convert.to_compare_safe(column)
        sort_values = sort_values.set_column(i, column_name, column._arrow_data)

    indices = _arrow_sort_values_to_indices(sort_values, col_dirs)
    return ArrowWeaveList(
        pc.take(self._arrow_data, indices), self.object_type, self._artifact
    )


@op(
    name="ArrowWeaveList-filter",
    input_type={
        "self": ArrowWeaveListType(),
        "filter_fn": lambda input_types: types.Function(
            {"row": input_types["self"].object_type, "index": types.Int()},
            types.optional(types.Boolean()),
        ),
    },
    output_type=lambda input_types: input_types["self"],
)
def filter(self, filter_fn):
    mask = _apply_fn_node_with_tag_pushdown(self, filter_fn)
    arrow_mask = mask._arrow_data_asarray_no_tags()
    arrow_mask = pc.fill_null(arrow_mask.cast(pa.bool_()), False)
    return ArrowWeaveList(
        arrow_as_array(self._arrow_data).filter(arrow_mask),
        self.object_type,
        self._artifact,
    )


def awl_group_by_result_object_type(
    object_type: types.Type, _key: types.Type
) -> tagged_value_type.TaggedValueType:
    return tagged_value_type.TaggedValueType(
        types.TypedDict(
            {
                "groupKey": _key,
            }
        ),
        ArrowWeaveListType(object_type),
    )


def awl_group_by_result_type(
    object_type: types.Type, key_type: types.Type
) -> "ArrowWeaveListType":
    return ArrowWeaveListType(awl_group_by_result_object_type(object_type, key_type))


@op(
    name="ArrowWeaveList-groupby",
    input_type={
        "self": ArrowWeaveListType(),
        "group_by_fn": lambda input_types: types.Function(
            {"row": input_types["self"].object_type}, types.Any()
        ),
    },
    output_type=lambda input_types: awl_group_by_result_type(
        input_types["self"].object_type, input_types["group_by_fn"].output_type
    ),
)
def groupby(self, group_by_fn):
    table = self._arrow_data
    unsafe_group_table_awl = _apply_fn_node_with_tag_pushdown(self, group_by_fn)
    group_table_awl = to_compare_safe(unsafe_group_table_awl.without_tags())
    group_table_as_array_awl_stripped = group_table_awl._arrow_data
    group_table_chunked = pa.chunked_array(
        pa.StructArray.from_arrays(
            [
                group_table_as_array_awl_stripped,
            ],
            names=["group_key"],
        )
    )
    group_table_chunked_unzipped = unzip_struct_array(group_table_chunked)
    group_cols = group_table_chunked_unzipped.column_names

    # Serializing a large arrow table and then reading it back
    # causes it to come back with more than 1 chunk. It seems the aggregation
    # operations don't like this. It will raise a cryptic error about
    # ExecBatches need to have the same link without this combine_chunks line
    # But combine_chunks doesn't seem like the most efficient thing to do
    # either, since it'll have to concatenate everything together.
    # But this fixes the crash for now!
    # TODO: investigate this as we optimize the arrow implementation
    group_table_combined = group_table_chunked_unzipped.combine_chunks()

    group_table_combined_indexed = group_table_combined.append_column(
        "_index", pa.array(np.arange(len(group_table_combined)))
    )
    awl_grouped = group_table_combined_indexed.group_by(group_cols)
    awl_grouped_agg = awl_grouped.aggregate([("_index", "list")])
    awl_grouped_agg_struct = unflatten_structs_in_flattened_table(awl_grouped_agg)

    combined = awl_grouped_agg_struct.column("_index_list").combine_chunks()
    val_lengths = combined.value_lengths()
    flattened_indexes = combined.flatten()
    values = arrow_as_array(table).take(flattened_indexes)
    offsets = np.cumsum(np.concatenate(([0], val_lengths)))
    grouped_results = pa.ListArray.from_arrays(offsets, values)
    grouped_awl = ArrowWeaveList(
        grouped_results, ArrowWeaveListType(self.object_type), self._artifact
    )
    effective_group_key_indexes = flattened_indexes.take(
        pa.array(offsets.tolist()[:-1]).cast(pa.int64())
    )
    effective_group_keys = arrow_as_array(unsafe_group_table_awl._arrow_data).take(
        effective_group_key_indexes
    )
    nested_effective_group_keys = pa.StructArray.from_arrays(
        [effective_group_keys], names=["groupKey"]
    )

    return arrow_tags.awl_add_arrow_tags(
        grouped_awl,
        nested_effective_group_keys,
        types.TypedDict({"groupKey": unsafe_group_table_awl.object_type}),
    )


@op(
    name="ArrowWeaveList-dropna",
    input_type={"self": ArrowWeaveListType()},
    output_type=lambda input_types: ArrowWeaveListType(
        types.non_none(input_types["self"].object_type)
    ),
)
def dropna(self):
    res = pc.drop_null(self._arrow_data)
    return ArrowWeaveList(res, types.non_none(self.object_type), self._artifact)


@op(name="ArrowWeaveList-count")
def count(self: ArrowWeaveList) -> int:
    return self._count()


@op(
    name="ArrowWeaveList-__getitem__",
    output_type=lambda input_types: primitive_list.getitem_output_type(
        {"arr": input_types["self"], "index": input_types["index"]},
        list_type=ArrowWeaveListType,
    ),
)
def index(
    self: ArrowWeaveList,
    index: typing.Optional[typing.Union[int, typing.List[typing.Optional[int]]]],
):
    return self._index(index)


@op(name="ArrowWeaveList-offset", output_type=lambda input_types: input_types["self"])
def offset(self: ArrowWeaveList, offset: int):
    return ArrowWeaveList(
        self._arrow_data.slice(offset), self.object_type, self._artifact
    )


@op(name="ArrowWeaveList-limit", output_type=lambda input_types: input_types["self"])
def limit(self: ArrowWeaveList, limit: int):
    return self._limit(limit)


@op(
    name="ArrowWeaveList-unnest",
    input_type={"self": ArrowWeaveListType(types.TypedDict({}))},
    output_type=lambda input_types: ArrowWeaveListType(
        types.TypedDict(
            {
                k: v
                if not (types.is_list_like(v) or isinstance(v, ArrowWeaveListType))
                else v.object_type
                for (k, v) in input_types["self"].object_type.property_types.items()
            }
        )
    ),
)
def unnest(self):
    if not self or not isinstance(self.object_type, types.TypedDict):
        return self

    list_cols = []
    new_obj_prop_types = {}
    for k, v_type in self.object_type.property_types.items():
        if types.is_list_like(v_type):
            list_cols.append(k)
            new_obj_prop_types[k] = op_def.normalize_type(v_type.object_type)
        else:
            new_obj_prop_types[k] = v_type
    if not list_cols:
        return self

    if isinstance(self._arrow_data, pa.StructArray):
        rb = pa.RecordBatch.from_struct_array(
            self._arrow_data
        )  # this pivots to columnar layout
        arrow_obj = pa.Table.from_batches([rb])
    else:
        arrow_obj = self._arrow_data

    # Split out tagged list columns into separate columns
    tag_col_name_map = {}
    for col in list_cols:
        col_data = arrow_obj.column(col).combine_chunks()
        if isinstance(
            self.object_type.property_types[col], tagged_value_type.TaggedValueType
        ):
            tag_col_name = f"{col}__tag__"
            tag_col_name_map[col] = tag_col_name

            col_tags = col_data.field("_tag")
            col_values = col_data.field("_value")

            arrow_obj = arrow_obj.remove_column(arrow_obj.column_names.index(col))
            arrow_obj = arrow_obj.append_column(tag_col_name, col_tags)
            arrow_obj = arrow_obj.append_column(col, col_values)

    # todo: make this more efficient. we shouldn't have to convert back and forth
    # from the arrow in-memory representation to pandas just to call the explode
    # function. but there is no native pyarrow implementation of this

    # This can replace int64 with float64! I believe it happens when we have nullable
    # int. pandas represents nullability using nan.
    # This requires us a hack in our AWL.validate function to allow float64 when
    # the weave type is int.
    # TODO: write an arrow based implementation, and remove the hack in validate.
    exploded_table = pa.Table.from_pandas(
        df=arrow_obj.to_pandas().explode(list_cols), preserve_index=False
    )

    # Reconstruct the tagged list columns
    for col in exploded_table.column_names:
        if col in tag_col_name_map:
            tag_col_name = tag_col_name_map[col]
            val_col = exploded_table.column(col).combine_chunks()
            tag_col = exploded_table.column(tag_col_name).combine_chunks()
            combined_col = arrow_tags.direct_add_arrow_tags(val_col, tag_col)
            exploded_table = exploded_table.remove_column(
                exploded_table.column_names.index(tag_col_name)
            )
            exploded_table = exploded_table.remove_column(
                exploded_table.column_names.index(col)
            )
            exploded_table = exploded_table.append_column(col, combined_col)

    return ArrowWeaveList(
        exploded_table,
        types.TypedDict(new_obj_prop_types),
        self._artifact,
    )


def _concat_output_type(input_types: typing.Dict[str, types.List]) -> types.Type:
    arr_type: types.List = input_types["arr"]
    inner_type = types.non_none(arr_type.object_type)

    if isinstance(inner_type, types.UnionType):
        if not all(types.is_list_like(t) for t in inner_type.members):
            raise ValueError(
                "opConcat: expected all members of inner type to be list-like"
            )

        new_union_members = [
            typing.cast(
                ArrowWeaveListType,
                tagged_value_type_helpers.push_down_tags_from_container_type_to_element_type(
                    t
                ),
            ).object_type
            for t in inner_type.members
        ]

        # merge the types into a single type
        new_inner_type = new_union_members[0]
        for t in new_union_members[1:]:
            new_inner_type = types.merge_types(new_inner_type, t)

        return ArrowWeaveListType(new_inner_type)

    elif isinstance(inner_type, tagged_value_type.TaggedValueType):
        inner_type_value = inner_type.value
        inner_type_tag = inner_type.tag
        inner_type_value_inner_type = typing.cast(
            ArrowWeaveListType, inner_type_value
        ).object_type

        return ArrowWeaveListType(
            tagged_value_type.TaggedValueType(
                inner_type_tag, inner_type_value_inner_type
            )
        )

    return inner_type


@op(
    name="ArrowWeaveList-concat",
    input_type={
        "arr": types.List(
            types.union(types.NoneType(), ArrowWeaveListType(types.Any()))
        )
    },
    output_type=_concat_output_type,
)
def concat(arr):
    arr = [item for item in arr if item != None]

    if len(arr) == 0:
        return convert.to_arrow([])
    elif len(arr) == 1:
        return arrow_tags.pushdown_list_tags(arr[0])

    tagged = list(builtin_map(lambda x: arrow_tags.pushdown_list_tags(x), arr))

    # We merge the lists mergesort-style, which is `O(n*log(n))`
    # DO NOT merge the lists reduce-style, which is `O(n^2)`
    return merge_concat(tagged)


def merge_concat(arr: list[ArrowWeaveList]) -> ArrowWeaveList:
    if len(arr) == 0:
        raise ValueError("arr must not be empty")
    if len(arr) == 1:
        return arr[0]
    left, right = merge_concat_split(arr)
    return merge_concat(left).concat(merge_concat(right))


def merge_concat_split(
    arr: list[ArrowWeaveList],
) -> tuple[list[ArrowWeaveList], list[ArrowWeaveList]]:
    if len(arr) < 2:
        raise ValueError("arr must have length of at least 2")
    middle_index = len(arr) // 2
    return arr[:middle_index], arr[middle_index:]


# # Putting this here instead of in number b/c it is just a map function
# bin_type = types.TypedDict({"start": types.Number(), "stop": types.Number()})


# @op(
#     name="ArrowWeaveListNumber-bin",
#     input_type={
#         "val": ArrowWeaveListType(types.optional(types.Number())),
#         "binFn": types.Function(
#             {"row": types.Number()},
#             types.TypedDict({"start": types.Number(), "stop": types.Number()}),
#         ),
#     },
#     output_type=ArrowWeaveListType(types.optional(bin_type)),
# )
# def number_bin(val, binFn):
#     tagged_awl = pushdown_list_tags(val)
#     res = _apply_fn_node(tagged_awl, binFn)
#     return res


def awl_object_type_with_index(object_type):
    return tagged_value_type.TaggedValueType(
        types.TypedDict({"indexCheckpoint": types.Int()}), object_type
    )


def arrow_weave_list_createindexcheckpoint_output_type(input_type):
    return ArrowWeaveListType(awl_object_type_with_index(input_type["arr"].object_type))


@op(
    name="arrowWeaveList-createIndexCheckpointTag",
    input_type={"arr": ArrowWeaveListType(types.Any())},
    output_type=arrow_weave_list_createindexcheckpoint_output_type,
)
def arrow_weave_list_createindexCheckpoint(arr):
    # Tags are always stored as a list of dicts, even if there is only one tag
    tag_array = pa.StructArray.from_arrays(
        [pa.array(np.arange(len(arr._arrow_data)))],
        names=["indexCheckpoint"],
    )
    return arrow_tags.awl_add_arrow_tags(
        arr,
        tag_array,
        types.TypedDict({"indexCheckpoint": types.Int()}),
    )


def vectorized_list_output_type(input_types):
    element_types = vectorized_input_types(input_types).values()
    return ArrowWeaveListType(types.List(types.union(*element_types)))


@op(
    name="ArrowWeaveList-vectorizedList",
    input_type=op_args.OpVarArgs(types.Any()),
    output_type=vectorized_list_output_type,
    render_info={"type": "function"},
)
def arrow_list_(**e):
    if len(e) == 0:
        return ArrowWeaveList(pa.nones(0), types.UnknownType(), None)
    res = vectorized_container_constructor_preprocessor(e)

    # Use our ArrowWeaveList concat implementation. Its guaranteed to
    # work for all possible AWL types.
    awls = [
        ArrowWeaveList(arr, object_type, None)
        for arr, object_type in zip(res.arrays, res.prop_types.values())
    ]
    result_values = awls[0]
    for next_awl in awls[1:]:
        result_values = result_values.concat(next_awl)

    concatted = result_values._arrow_data

    if len(concatted) == 0:
        values = concatted
    else:
        take_ndxs = []
        for row_ndx in range(res.max_len):
            take_ndxs.extend([row_ndx + i * res.max_len for i in range(len(e))])
        values = concatted.take(pa.array(take_ndxs))
    offsets = pa.array(
        [i * len(e) for i in range(res.max_len)] + [res.max_len * len(e)]
    )
    return ArrowWeaveList(
        pa.ListArray.from_arrays(offsets, values),
        types.List(result_values.object_type),
        res.artifact,
    )


def vectorized_arrow_pick_output_type(input_types):
    inner_type = types.union(*list(input_types["self"].property_types.values()))
    if isinstance(input_types["key"], ArrowWeaveListType):
        return ArrowWeaveListType(inner_type)
    else:
        return types.List(inner_type)


# Combining both list and arrow into the same op since dispatch operates on the
# first input, and the difference would only be in the second input
@op(
    name="_vectorized_list_like-pick",
    input_type={
        "self": types.TypedDict({}),
        "key": types.union(
            ArrowWeaveListType(types.optional(types.String())),
            types.List(types.optional(types.String())),
        ),
    },
    output_type=vectorized_arrow_pick_output_type,
)
def vectorized_arrow_pick(self, key):
    if isinstance(key, list):
        return [self.get(k, None) if k != None else None for k in key]
    de = arrow_as_array(key._arrow_data_asarray_no_tags()).dictionary_encode()
    new_dictionary = pa.array(
        [self.get(key, None) for key in de.dictionary.to_pylist()]
    )
    new_indices = de.indices
    new_array = pa.DictionaryArray.from_arrays(new_indices, new_dictionary)
    result = new_array.dictionary_decode()
    return ArrowWeaveList(
        result,
    )


@arrow_op(
    name="ArrowWeaveList-vectorizedIsNone",
    input_type={"self": ArrowWeaveListType(types.optional(types.Any()))},
    output_type=ArrowWeaveListType(types.Boolean()),
)
def vectorized_is_none(self):
    # Need to break out to python due to this issue we reported
    # https://github.com/apache/arrow/issues/34315
    # TODO: Remove this once the issue is fixed
    # Open PR here: https://github.com/apache/arrow/pull/34408
    if isinstance(self._arrow_data, pa.UnionArray):
        return ArrowWeaveList(
            pa.array([x == None for x in self._arrow_data.to_pylist()]),
            types.Boolean(),
            self._artifact,
        )
    return ArrowWeaveList(
        self._arrow_data.is_null(),
        types.Boolean(),
        self._artifact,
    )


@op(
    name="ArrowWeaveList-randomlyDownsample",
    input_type={
        "self": ArrowWeaveListType(),
        "n": types.Int(),
    },
    output_type=lambda input_types: input_types["self"],
    pure=False,
)
def sample(self, n):
    if n < 0:
        raise ValueError("n must be non-negative")

    elif n >= len(self):
        return self

    arr = np.zeros(len(self), dtype=bool)
    indices = np.random.choice(len(self), n, replace=False)
    arr[indices] = True

    arrow_data = self._arrow_data
    new_arrow_data = arrow_data.filter(pa.array(arr))

    return ArrowWeaveList(
        new_arrow_data,
        self.object_type,
        self._artifact,
    )


def flatten_return_object_type(object_type):
    if types.is_list_like(object_type):
        object_type = object_type.object_type
    return object_type


def flatten_return_type(input_types):
    return ArrowWeaveListType(
        flatten_return_object_type(input_types["arr"].object_type)
    )


@op(
    name="ArrowWeaveList-flatten",
    input_type={"arr": ArrowWeaveListType()},
    output_type=flatten_return_type,
)
def flatten(arr):
    # TODO:
    #   - handle N levels instead of 1
    #   - handle tags
    arrow_data = arr._arrow_data
    if is_list_arrowweavelist(arr):
        arrow_data = arrow_data.flatten()
    return ArrowWeaveList(
        arrow_data,
        flatten_return_object_type(arr.object_type),
        arr._artifact,
    )



# Deserialize and serialize have to match the javascript implementation, which
# needlessly adds an entry for nodes and ops. It'd be simpler to just encode
# nodes and inline the ops in their respective output nodes.

import typing
import hashlib
import json
import random

from . import value_or_error


from . import graph
from . import weave_types as types
from . import errors
from . import weave_internal
from . import storage
from . import memo


NodeOrOp = typing.Union[graph.Node, graph.Op]

# TODO(dg): Replace Any
SerializedNode = typing.Any
MapNodeOrOpToSerialized = typing.Dict[NodeOrOp, typing.Tuple[SerializedNode, int]]


class SerializedReturnType(typing.TypedDict):
    nodes: typing.List[SerializedNode]
    targetNodes: typing.List[int]


def _serialize_node(node: NodeOrOp, serialized_nodes: MapNodeOrOpToSerialized) -> int:
    if node in serialized_nodes:
        return serialized_nodes[node][1]

    if isinstance(node, graph.Op):
        param_indexes = {}
        for param_name, param_node in node.inputs.items():
            param_indexes[param_name] = _serialize_node(param_node, serialized_nodes)
        op_id = len(serialized_nodes)
        serialized_nodes[node] = ({"name": node.name, "inputs": param_indexes}, op_id)
        return op_id
    if isinstance(node, graph.ConstNode):
        if isinstance(node.type, types.Function):
            op_id = _serialize_node(node.val.from_op, serialized_nodes)
            node_id = len(serialized_nodes)
            serialized_nodes[node] = (
                {
                    "nodeType": "const",
                    "type": node.type.to_dict(),
                    "val": {"nodeType": "output", "fromOp": op_id},
                },
                node_id,
            )
            return node_id
        else:
            node_id = len(serialized_nodes)
            serialized_nodes[node] = (
                node.to_json(),
                node_id,
            )
            return node_id
    elif isinstance(node, graph.OutputNode):
        op_id = _serialize_node(node.from_op, serialized_nodes)
        node_id = len(serialized_nodes)
        serialized_nodes[node] = (
            {"nodeType": "output", "type": node.type.to_dict(), "fromOp": op_id},
            node_id,
        )
        return node_id
    elif isinstance(node, graph.VarNode):
        node_id = len(serialized_nodes)
        serialized_nodes[node] = (node.to_json(), node_id)
        return node_id

    raise ValueError(f"Could not serialize node: {node}")


def serialize(graphs: typing.List[graph.Node]) -> SerializedReturnType:
    serialized: MapNodeOrOpToSerialized = {}
    target_nodes = [_serialize_node(n, serialized) for n in graphs]
    nodes = list(range(len(serialized)))
    for node_json, node_id in serialized.values():
        nodes[node_id] = node_json
    return {"nodes": nodes, "targetNodes": target_nodes}


def _is_lambda(node: graph.Node):
    return isinstance(node, graph.ConstNode) and (
        isinstance(node.val, graph.OutputNode) or isinstance(node.val, graph.VarNode)
    )


@memo.memo
def node_id(node: graph.Node):
    hashable = {"type": node.type.to_dict()}
    if isinstance(node, graph.OutputNode):
        hashable["op_name"] = node.from_op.name
        hashable["inputs"] = {
            arg_name: node_id(arg_node)
            if not _is_lambda(arg_node)
            else json.dumps(arg_node.to_json())
            for arg_name, arg_node in node.from_op.inputs.items()
        }
    elif isinstance(node, graph.VarNode):
        # Ensure we don't share var nodes. That's invalid!
        hashable["name"] = str(random.random())
    elif isinstance(node, graph.ConstNode):
        if isinstance(node.val, graph.OutputNode) or isinstance(
            node.val, graph.VarNode
        ):
            # Ensure we don't share function nodes. That's invalid!
            hashable["val"] = str(random.random())
        else:
            hashable["val"] = storage.to_python(node.val)
    else:
        raise errors.WeaveInternalError("invalid node encountered: %s" % node)
    hash = hashlib.md5()
    hash.update(json.dumps(hashable).encode())
    return hash.hexdigest()


def _deserialize_node(
    index: int,
    nodes: typing.List[SerializedNode],
    parsed_nodes: typing.MutableMapping[int, graph.Node],
    hashed_nodes: typing.MutableMapping[str, graph.Node],
) -> graph.Node:
    if index in parsed_nodes:
        return parsed_nodes[index]
    node = nodes[index]

    parsed_node: graph.Node
    if node["nodeType"] == "const":
        if isinstance(node["type"], dict) and node["type"]["type"] == "function":
            fn_body_node = node["val"]
            parsed_fn_body_node: graph.Node
            if fn_body_node["nodeType"] == "var":
                parsed_fn_body_node = weave_internal.make_var_node(
                    types.TypeRegistry.type_from_dict(fn_body_node["type"]),
                    fn_body_node["varName"],
                )
            elif fn_body_node["nodeType"] == "const":
                fn_body_const_val = fn_body_node["val"]
                if (
                    isinstance(fn_body_const_val, dict)
                    and "nodeType" in fn_body_const_val
                ):
                    # This case happens when we have a quoted function.
                    fn_body_const_val = graph.Node.node_from_json(fn_body_const_val)
                parsed_fn_body_node = weave_internal.make_const_node(
                    types.TypeRegistry.type_from_dict(fn_body_node["type"]),
                    fn_body_const_val,
                )
            elif fn_body_node["nodeType"] == "output":
                op = nodes[fn_body_node["fromOp"]]
                params = {}
                for param_name, param_node_index in op["inputs"].items():
                    params[param_name] = _deserialize_node(
                        param_node_index, nodes, parsed_nodes, hashed_nodes
                    )
                node_type = types.TypeRegistry.type_from_dict(node["type"])
                if not isinstance(node_type, types.Function):
                    raise errors.WeaveInternalError(
                        "expected function type, got %s" % node_type
                    )
                parsed_fn_body_node = weave_internal.make_output_node(
                    node_type.output_type,
                    op["name"],
                    params,
                )
            else:
                raise errors.WeaveInternalError(
                    "invalid function node encountered in deserialize"
                )
            parsed_node = graph.ConstNode(
                types.TypeRegistry.type_from_dict(node["type"]), parsed_fn_body_node
            )
        else:
            parsed_node = graph.ConstNode.from_json(node)
    elif node["nodeType"] == "output":
        op = nodes[node["fromOp"]]
        params = {}
        for param_name, param_node_index in op["inputs"].items():
            params[param_name] = _deserialize_node(
                param_node_index, nodes, parsed_nodes, hashed_nodes
            )
        parsed_node = graph.OutputNode(
            types.TypeRegistry.type_from_dict(node["type"]), op["name"], params
        )
    elif node["nodeType"] == "var":
        parsed_node = graph.VarNode.from_json(node)
    id_ = node_id(parsed_node)
    if id_ in hashed_nodes:
        parsed_node = hashed_nodes[id_]
    else:
        hashed_nodes[id_] = parsed_node
    parsed_nodes[index] = parsed_node
    return parsed_node


def deserialize(
    serialized: SerializedReturnType,
) -> value_or_error.ValueOrErrors[graph.Node]:
    # For some reason what javascript sends isn't guaranteed to be sorted
    # along the graph topology. If it were we could do an easy linear
    # implementation. But its not so we recurse for now.
    nodes = serialized["nodes"]
    target_nodes = serialized.get("targetNodes")
    if target_nodes is None:
        # Handle old format for backward compat with request replay datasets.
        target_nodes = serialized["rootNodes"]  # type: ignore

    parsed_nodes: dict[int, graph.Node] = {}
    # WeaveJS does not do a good job deduplicating nodes currently, so we do it here.
    # This ensures we don't execute the same node many times.
    hashed_nodes: dict[str, graph.Node] = {}

    target_node_values = value_or_error.ValueOrErrors.from_values(target_nodes)

    with memo.memo_storage():
        return target_node_values.safe_map(
            lambda i: _deserialize_node(i, nodes, parsed_nodes, hashed_nodes)
        )



from ..compile_domain import wb_gql_op_plugin
from ..api import op
from .. import weave_types as types
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    gql_root_op,
)

# Section 1/6: Tag Getters
# None

# Section 2/6: Root Ops
# None

# Section 3/6: Attribute Getters
gql_prop_op("artifactAlias-alias", wdt.ArtifactAliasType, "alias", types.String())

# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "artifactAlias-artifact",
    wdt.ArtifactAliasType,
    "artifactCollection",
    wdt.ArtifactCollectionType,
)

# Section 5/6: Connection Ops
# None

# Section 6/6: Non Standard Business Logic Ops
# None



"""
This module provides utilities for handling errors in the Weave server. It provides
3 main things:
1. WeaveInternalHttpException - a subclass of werkzeug.exceptions.HTTPException that can
   be used to create HTTP Exceptions directly from an error code. This replaces the old
   errors.WeaveHTTPException. Importantly, it is a subless of werkzeug.exceptions.HTTPException
   which means that flask will properly handle it and return the correct HTTP response. Currently,
   it is used by the I/O service when encountering an HTTP error, and weave_http when getting
   a download error. It is also used by the `client_safe_http_exceptions_as_werkzeug` to mask
   non-client-safe errors with a generic 500 error.
2. client_safe_http_exceptions_as_werkzeug - a context manager that can be used to ensure any
    error that occurs within the context is properly transformed into a werkzeug exception
    and that the error is client-safe - this is used by middleware and execute methods on the server
3. maybe_extract_code_from_exception - a method that can be used to extract the HTTP code from
    various library's exceptions. This is used by the `client_safe_http_exceptions_as_werkzeug`
    as well as internally by i.o service to extract the HTTP code from the error. This is needed
    because we have different libraries under the hood that produce different exception classes.
"""


from contextlib import contextmanager
import logging
import typing
import gql
import requests

from werkzeug import Response, exceptions as werkzeug_exceptions, http as werkzeug_http


from . import errors
from . import util


class WeaveInternalHttpException(werkzeug_exceptions.HTTPException):
    @classmethod
    def from_code(
        cls, code: typing.Optional[int] = None, description: typing.Optional[str] = None
    ) -> "WeaveInternalHttpException":
        if code is None:
            code = 500
        if description is None:
            description = werkzeug_http.HTTP_STATUS_CODES.get(code, "Unknown Error")
        res = cls(description, Response(None, code))
        res.code = code
        return res


@contextmanager
def client_safe_http_exceptions_as_werkzeug() -> typing.Generator[None, None, None]:
    try:
        yield
    except Exception as e:
        code = maybe_extract_code_from_exception(e)
        if code is None:
            raise
        # If the exception is a errors.WeaveBadRequest, then send up the 400
        # code. This is the one case we want to pass along 400s
        if code in _client_safe_code_allowlist or (
            isinstance(e, errors.WeaveBadRequest) and code == 400
        ):
            werkzeug_exceptions.abort(code)
        else:
            # Here, we mask the original exception with a generic 500 error.
            raise WeaveInternalHttpException.from_code(500)


def maybe_extract_code_from_exception(e: Exception) -> typing.Optional[int]:
    """
    Use this method to extract the HTTP codes from various library's exceptions.
    """
    if isinstance(e, werkzeug_exceptions.HTTPException):
        return _extract_code_from_werkzeug_http_exception(e)
    elif isinstance(e, requests.exceptions.RequestException):
        return _extract_code_from_request_lib_request_exception(e)
    elif isinstance(e, gql.transport.exceptions.TransportServerError):
        return _extract_code_from_gql_lib_error(e)
    elif isinstance(e, errors.WeaveBadRequest):
        return _extract_code_from_weave_bad_request_error(e)
    return None


def _extract_code_from_werkzeug_http_exception(
    e: werkzeug_exceptions.HTTPException,
) -> typing.Optional[int]:
    if isinstance(e.code, int):
        return e.code
    return None


def _extract_code_from_request_lib_request_exception(
    e: requests.exceptions.RequestException,
) -> typing.Optional[int]:
    if isinstance(e, requests.HTTPError):
        if e.response is not None and isinstance(e.response.status_code, int):
            return e.response.status_code
    elif isinstance(e, requests.exceptions.ReadTimeout):
        # Convert internal read timeout to 502 (Bad Gateway)
        util.capture_exception_with_sentry_if_available(
            e, ("requests.exceptions.ReadTimeout",)
        )
        logging.warning(f"Converting requests.exceptions.ReadTimeout to 502")
        return 502
    return None


def _extract_code_from_gql_lib_error(
    e: gql.transport.exceptions.TransportServerError,
) -> typing.Optional[int]:
    if isinstance(e.code, int):
        return e.code
    return None


def _extract_code_from_weave_bad_request_error(
    e: errors.WeaveBadRequest,
) -> typing.Optional[int]:
    message = str(e.args[0]) if len(e.args) > 0 else ""
    util.capture_exception_with_sentry_if_available(e, (message,))
    logging.warning(f"Converting WeaveBadRequest to 400: {message}")
    return 400


_client_safe_code_allowlist = {
    502,  # Bad Gateway
    503,  # Service Unavailable
    504,  # Gateway Timeout
    429,  # Too Many Requests
    # Comment from CVP: https://github.com/wandb/weave-internal/pull/679#discussion_r1136214697
    # We may want to include the codes below (they are included in artifact code in sdk). Leaving
    # commented for now until we decide on if the client should handle these
    # 308,  # Resume Incomplete (only for Google Cloud Storage uploads, might not be needed)
    # 408,  # Request timeout
    # 409,  # Conflictt
}



import dataclasses
import typing

from . import weave_types as types


@dataclasses.dataclass
class Run:
    id: str
    op_name: str
    state: str = dataclasses.field(default_factory=lambda: "pending")
    prints: list[str] = dataclasses.field(default_factory=list)
    history: list[dict] = dataclasses.field(default_factory=list)
    inputs: dict[str, typing.Any] = dataclasses.field(default_factory=dict)
    output: typing.Any = dataclasses.field(default_factory=lambda: None)


types.RunType.instance_classes = Run
types.RunType.instance_class = Run



# This is essentially a transaction manager and we may want to rename it.
# You can create it, then any mutations you make to objects within the context
# will be collected, and only written when the context exits.
# Can be used with weave_api.finish() to batch mutations from user code.
#
# Its experimental for now.

import contextvars
import typing
import contextlib
import dataclasses
import typing
from typing import Any

if typing.TYPE_CHECKING:
    from . import ref_base

from . import weave_types as types
from . import box
from . import uris


@dataclasses.dataclass
class MutationRecord:
    mutation_name: str
    args: typing.List[typing.Any]


@dataclasses.dataclass
class ObjectRecord:
    val: typing.Any
    type: types.Type
    branched_from_uri: typing.Optional[str] = None
    mutations: typing.List[MutationRecord] = dataclasses.field(default_factory=list)


class NoResult:
    pass


@dataclasses.dataclass
class ObjectContext:
    objects: dict[str, ObjectRecord] = dataclasses.field(default_factory=dict)

    def add_ref(self, target_uri: str, val: typing.Any, type: types.Type) -> None:
        self.objects[target_uri] = ObjectRecord(val, type)

    def lookup_ref_val(self, target_uri: str) -> typing.Any:
        if target_uri not in self.objects:
            return NoResult()
        return self.objects[target_uri].val

    def lookup_ref_type(self, target_uri: str) -> typing.Union[types.Type, NoResult]:
        if target_uri not in self.objects:
            return NoResult()
        return self.objects[target_uri].type

    def add_mutation(
        self,
        target_uri: str,
        source_uri: str,
        new_val: typing.Any,
        make_new_type: typing.Callable[[types.Type], types.Type],
        mutation: typing.Any,
    ) -> None:
        branched_from_uri = None
        if source_uri != target_uri:
            branched_from_uri = source_uri

        source_record = self.objects.get(source_uri)
        if branched_from_uri is not None and source_record is None:
            raise ValueError("Can't branch object that doesn't exist")

        target_record = self.objects.get(target_uri)
        if target_record is None:
            self.objects[target_uri] = ObjectRecord(
                new_val,
                make_new_type(types.UnknownType()),
                branched_from_uri,
                mutations=[mutation],
            )
        else:
            self.objects[target_uri].val = new_val
            self.objects[target_uri].type = make_new_type(target_record.type)
            if branched_from_uri is not None:
                # we're overwriting a target with a new branch, so reset
                # mutations
                self.objects[target_uri].mutations = [mutation]
            else:
                self.objects[target_uri].mutations.append(mutation)

    def finish_mutation(self, target_uri: str) -> None:
        from . import artifact_fs, artifact_local, artifact_wandb

        target_record = self.objects.get(target_uri)
        if target_record is None:
            raise ValueError("No object to finish mutation on")
        if not target_record.mutations:
            return

        if target_record.branched_from_uri is None:
            source_uri = uris.WeaveURI.parse(target_uri)
            art = artifact_local.LocalArtifact(source_uri.name, source_uri.version)
        else:
            source_uri = uris.WeaveURI.parse(target_record.branched_from_uri)
            # Note: the mutation bookeeping is done at the individual object level,
            # so the path component is often non-empty. However, since we are managing
            # state at the artifact level, we need to ensure the path is empty to conform
            # to the expectations of the branching logic.
            source_uri.path = ""  # type: ignore
            art = artifact_local.LocalArtifact.fork_from_uri(source_uri)  # type: ignore

        obj = box.box(target_record.val)
        target_branch = uris.WeaveURI.parse(target_uri).version
        if target_branch is None:
            raise ValueError("No branch to finish mutation on")

        from weave import artifact_wandb

        # Hack: if the target branch looks like a commit hash, then we
        # don't want to use it as a branch - this is the case that we are
        # mutating an artifact directly without a branch. Seems like we
        # might want to put this logic elsewhere, but for now this works.
        if artifact_wandb.likely_commit_hash(target_branch):
            target_branch = None
        weave_type = types.TypeRegistry.type_of(obj)
        ref = art.set("obj", weave_type, obj)
        artifact_fs.update_weave_meta(weave_type, art)
        art.save(branch=target_branch)  # type: ignore

    def finish_mutations(self) -> None:
        for target_uri in self.objects.keys():
            self.finish_mutation(target_uri)


_object_context: contextvars.ContextVar[
    typing.Optional[ObjectContext]
] = contextvars.ContextVar("_object_context", default=None)


@contextlib.contextmanager
def object_context() -> typing.Iterator[ObjectContext]:
    ctx = _object_context.get()
    token = None
    if ctx is None:
        ctx = ObjectContext()
        token = _object_context.set(ctx)
    try:
        yield ctx
    finally:
        if token is not None:
            _object_context.reset(token)
            ctx.finish_mutations()


def get_object_context() -> typing.Optional[ObjectContext]:
    return _object_context.get()



import typing
from . import graph as _graph
from . import graph_mapper as _graph_mapper
from . import storage as _storage
from . import trace as _trace
from . import weave_internal as _weave_internal
from . import errors as _errors
from . import ops as _ops
from . import context as _context

# exposed as part of api
from . import weave_types as types
from . import types_numpy as _types_numpy
from . import errors
from .decorators import weave_class, op, mutation, type
from .op_args import OpVarArgs
from .op_def import OpDef
from . import usage_analytics
from .context import (
    use_fixed_server_port,
    use_frontend_devmode,
    # eager_execution,
    # lazy_execution,
)
from .server import capture_weave_server_logs
from .val_const import const
from .file_base import File, Dir
from .dispatch import RuntimeConstNode

from .weave_internal import define_fn

Node = _graph.Node


def save(node_or_obj, name=None):
    if isinstance(node_or_obj, _graph.Node):
        return _ops.save(node_or_obj, name=name)
    else:
        # If the user does not provide a branch, then we explicitly set it to
        # the default branch, "latest".
        branch = None
        name_contains_branch = name is not None and ":" in name
        if not name_contains_branch:
            branch = "latest"
        ref = _storage.save(node_or_obj, name=name, branch=branch)
        if name is None:
            # if the user didn't provide a name, the returned reference
            # will be to the specific version
            uri = ref.uri
        else:
            # otherwise the reference will be to whatever branch was provided
            # or the "latest" branch if only a name was provided.
            uri = ref.branch_uri
        return _ops.get(str(uri))


def publish(node_or_obj, name=None):
    if isinstance(node_or_obj, _graph.Node):
        node_or_obj = use(node_or_obj)

    ref = _storage.publish(node_or_obj, name)
    return _weave_internal.make_const_node(ref.type, ref.obj)


def get(ref_str):
    obj = _storage.get(ref_str)
    ref = _storage._get_ref(obj)
    return _weave_internal.make_const_node(ref.type, obj)


def use(nodes, client=None):
    usage_analytics.use_called()
    if client is None:
        client = _context.get_client()
    return _weave_internal.use(nodes, client)


def _get_ref(obj):
    if isinstance(obj, _storage.Ref):
        return obj
    ref = _storage.get_ref(obj)
    if ref is None:
        raise _errors.WeaveApiError("obj is not a weave object: %s" % obj)
    return ref


def versions(obj):
    if isinstance(obj, _graph.ConstNode):
        obj = obj.val
    elif isinstance(obj, _graph.OutputNode):
        obj = use(obj)
    ref = _get_ref(obj)
    return ref.versions()


def expr(obj):
    ref = _get_ref(obj)
    return _trace.get_obj_expr(ref)


def type_of(obj: typing.Any) -> types.Type:
    return types.TypeRegistry.type_of(obj)


def weave(obj: typing.Any) -> RuntimeConstNode:
    return _weave_internal.make_const_node(type_of(obj), obj)  # type: ignore


def from_pandas(df):
    return _ops.pandas_to_awl(df)
